import math

from .AnimationClip import AABB, PackedFloatVector, PackedIntVector
from .NamedObject import NamedObject
from .Texture2D import StreamingInfo
from ..helpers.ResourceReader import get_resource_data
from ..math import Matrix4x4, Vector3


class MinMaxAABB:
    def __init__(self, reader):
        self.m_Min = reader.read_vector3()
        self.m_Max = reader.read_vector3()


class CompressedMesh:
    def __init__(self, reader):
        version = reader.version
        self.m_Vertices = PackedFloatVector(reader)
        self.m_UV = PackedFloatVector(reader)
        if version[0] < 5:
            self.m_BindPoses = PackedFloatVector(reader)
        self.m_Normals = PackedFloatVector(reader)
        self.m_Tangents = PackedFloatVector(reader)
        self.m_Weights = PackedIntVector(reader)
        self.m_NormalSigns = PackedIntVector(reader)
        self.m_TangentSigns = PackedIntVector(reader)
        if version[0] >= 5:
            self.m_FloatColors = PackedFloatVector(reader)
        self.m_BoneIndices = PackedIntVector(reader)
        self.m_Triangles = PackedIntVector(reader)
        if version[0] > 3 or (version[0] == 3 and version[1] >= 5):  # 3.5 and up
            if version[0] < 5:
                self.m_Colors = PackedIntVector(reader)
            else:
                self.m_UVInfo = reader.read_u_int()


class StreamInfo:
    def __init__(self, **kwargs):
        if "reader" in kwargs:
            reader = kwargs["reader"]
            version = reader.version
            self.channelMask = reader.read_u_int()
            self.offset = reader.read_u_int()

            if version[0] < 4:  # 4.0 down
                self.stride = reader.read_u_int()
                self.align = reader.read_u_int()
            else:
                self.stride = reader.read_byte()
                self.dividerOp = reader.read_byte()
                self.frequency = reader.read_u_short()
        else:
            self.__dict__ = kwargs


class ChannelInfo:
    def __init__(self, reader):
        self.stream = reader.read_byte()
        self.offset = reader.read_byte()
        self.format = reader.read_byte()
        self.dimension = reader.read_byte()


class VertexData:
    def __init__(self, reader):
        self.reader = reader
        version = reader.version

        if version[0] < 2018:  # 2018 down
            self.m_CurrentChannels = reader.read_u_int()

        self.m_VertexCount = reader.read_u_int()

        if version[0] >= 4:  # 4.0 and up
            m_ChannelsSize = reader.read_int()
            self.m_Channels = [ChannelInfo(reader) for _ in range(m_ChannelsSize)]

        if version[0] < 5:  # 5.0 down
            if version[0] < 4:  #
                m_StreamsSize = 4
            else:
                m_StreamsSize = reader.read_int()

            self.m_Streams = [StreamInfo(reader=reader) for _ in range(m_StreamsSize)]

            if version[0] < 4:  # 4.0 down
                self.GetChannels()
        else:  # 5.0 and up
            self.GetStreams()

        self.m_DataSize = reader.read_bytes(reader.read_int())
        reader.align_stream()

    def GetStreams(self):
        streamCount = 1 + (
            0 if not self.m_Channels else max([x.stream for x in self.m_Channels])
        )
        self.m_Streams = {}
        offset = 0
        for s in range(streamCount):
            chnMask = 0
            stride = 0
            for chn, m_Channel in enumerate(self.m_Channels):
                if m_Channel.stream == s:
                    if m_Channel.dimension > 0:
                        chnMask |= 1 << chn  # Shift 1UInt << chn
                    stride += m_Channel.dimension * MeshHelper.GetChannelFormatSize(
                        m_Channel.format
                    )
            self.m_Streams[s] = StreamInfo(
                channelMask=chnMask,
                offset=offset,
                stride=stride,
                dividerOp=0,
                frequency=0,
            )
            offset += self.m_VertexCount * stride
            # static size_t align_streamSize (size_t size) { return (size + (kVertexStreamAlign-1)) & ~(kVertexStreamAlign-1)
            offset = (offset + (16 - 1)) & ~(
                16 - 1
            )  # (offset + (16u - 1u)) & ~(16u - 1u);

    def GetChannels(self):
        self.m_Channels = []  # ChannelInfo[6]
        for i in range(6):
            self.m_Channels.append(ChannelInfo(self.reader))
        for s, m_Stream in enumerate(self.m_Streams):
            channelMask = bytearray(m_Stream.channelMask)  # BitArray
            offset = 0
            for i in range(6):
                if channelMask[i]:
                    m_Channel = self.m_Channels[i]
                    m_Channel.stream = s
                    m_Channel.offset = offset
                    if i in [0, 1]:
                        # 0 - kShaderChannelVertex
                        # 1 - kShaderChannelNormal
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 3
                    elif i == 2:  # kShaderChannelColor
                        m_Channel.format = 2  # kChannelFormatColor
                        m_Channel.dimension = 4
                    elif i in [3, 4]:
                        # 3 - kShaderChannelTexCoord0
                        # 4 - kShaderChannelTexCoord1
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 2
                    elif i == 5:  # kShaderChannelTangent
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 4
                    offset += m_Channel.dimension * MeshHelper.GetChannelFormatSize(
                        m_Channel.format
                    )

    def FixChannel(self):
        if any(
            [x.dimension > 4 for x in self.m_Channels]
        ):  # m_Channels.FirstOrDefault(x => x.dimension > 4) != null: #
            fixStream = max(
                [x.stream for x in self.m_Channels]
            )  # m_Channels.Max(x => x.stream)
            fixChannels = [
                x for x in self.m_Channels if x.dimension > 0 and x.stream == fixStream
            ]
            stride = 0
            for i, curChannel in enumerate(fixChannels):
                preChannel = fixChannels[i - 1]
                offset = curChannel.offset - preChannel.offset
                preChannel.dimension = offset / MeshHelper.GetChannelFormatSize(
                    preChannel.format
                )
                stride += offset
            # Fix Last
            m_Channel = fixChannels[-1]
            streamSize = len(self.m_DataSize) - self.m_Streams[fixStream].offset
            totalStride = streamSize / self.m_VertexCount
            channelStride = totalStride - stride
            m_Channel.dimension = channelStride / MeshHelper.GetChannelFormatSize(
                m_Channel.format
            )
            self.GetStreams()


class BoneWeights4:
    def __init__(self, reader=None):
        if reader:
            self.weight = reader.read_float_array(4)
            self.boneIndex = reader.read_int_array(4)
        else:
            self.weight = [0.0] * 4
            self.boneIndex = [0] * 4


class BlendShapeVertex:
    def __init__(self, reader):
        self.vertex = reader.read_vector3()
        self.normal = reader.read_vector3()
        self.tangent = reader.read_vector3()
        self.index = reader.read_u_int()


class MeshBlendShape:
    def __init__(self, reader):
        version = reader.version

        if version[0] == 4 and version[1] < 3:  # 4.3 down
            self.name = reader.read_aligned_string()
        self.firstVertex = reader.read_u_int()
        self.vertexCount = reader.read_u_int()
        if version[0] == 4 and version[1] < 3:  # 4.3 down
            self.aabbMinDelta = reader.read_vector3()
            self.aabbMaxDelta = reader.read_vector3()
        self.hasNormals = reader.read_boolean()
        self.hasTangents = reader.read_boolean()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            reader.align_stream()


class MeshBlendShapeChannel:
    def __init__(self, reader):
        self.name = reader.read_aligned_string()
        self.nameHash = reader.read_u_int()
        self.frameIndex = reader.read_int()
        self.frameCount = reader.read_int()


class BlendShapeData:
    def __init__(self, reader):
        version = reader.version

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            numVerts = reader.read_int()
            self.vertices = [BlendShapeVertex(reader) for _ in range(numVerts)]

            numShapes = reader.read_int()
            self.shapes = [MeshBlendShape(reader) for _ in range(numShapes)]

            numChannels = reader.read_int()
            self.channels = [MeshBlendShapeChannel(reader) for _ in range(numChannels)]
            self.fullWeights = reader.read_float_array()
        else:
            m_ShapesSize = reader.read_int()
            self.m_Shapes = [MeshBlendShape(reader) for _ in range(m_ShapesSize)]
            reader.align_stream()
            m_ShapeVerticesSize = reader.read_int()
            self.m_ShapeVertices = [
                BlendShapeVertex(reader) for _ in range(m_ShapeVerticesSize)
            ]


class SubMesh:
    def __init__(self, reader):
        version = reader.version

        self.firstByte = reader.read_u_int()
        self.indexCount = reader.read_u_int()
        self.topology = reader.read_int()

        if version[0] < 4:  # 4.0 down
            self.triangleCount = reader.read_u_int()

        if version[0] > 2017 or (
            version[0] == 2017 and version[1] >= 3
        ):  # 2017.3 and up
            self.baseVertex = reader.read_u_int()

        if version[0] >= 3:  # 3.0 and up
            self.firstVertex = reader.read_u_int()
            self.vertexCount = reader.read_u_int()
            self.localAABB = AABB(reader)


class Mesh(NamedObject):
    m_StreamData: None

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = reader.version

        self.m_Use16BitIndices = True
        self.m_Indices = []

        if version[0] < 3 or (version[0] == 3 and version[1] < 5):  # 3.5 down
            self.m_Use16BitIndices = reader.read_int() > 0

        if version[0] == 2 and version[1] <= 5:  # 2.5 and down
            m_IndexBuffer_size = reader.read_int()

            if self.m_Use16BitIndices:  #
                self.m_IndexBuffer = [
                    reader.read_u_short()
                    for _ in range(math.ceil(m_IndexBuffer_size / 2))
                ]
                reader.align_stream()
            else:
                self.m_IndexBuffer = reader.read_u_int_array(m_IndexBuffer_size // 4)

        m_SubMeshesSize = reader.read_int()
        self.m_SubMeshes = [SubMesh(reader) for _ in range(m_SubMeshesSize)]

        if version[0] > 4 or (version[0] == 4 and version[1] >= 1):  # 4.1 and up
            self.m_Shapes = BlendShapeData(reader)

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_BindPose = reader.read_matrix_array()
            self.m_BoneNameHashes = reader.read_u_int_array()
            self.m_RootBoneNameHash = reader.read_u_int()

        if version[0] > 2 or (version[0] == 2 and version[1] >= 6):  # 2.6.0 and up
            if version[0] >= 2019:  # 2019 and up
                m_BonesAABBSize = reader.read_int()
                self.m_BonesAABB = [MinMaxAABB(reader) for _ in range(m_BonesAABBSize)]
                self.m_VariableBoneCountWeights = reader.read_u_int_array()

            self.m_MeshCompression = reader.read_byte()
            if version[0] >= 4:  #
                if version[0] < 5:  #
                    self.m_StreamCompression = reader.read_byte()
                self.m_IsReadable = reader.read_boolean()
                self.m_KeepVertices = reader.read_boolean()
                self.m_KeepIndices = reader.read_boolean()
            reader.align_stream()

            # Unity fixed it in 2017.3.1p1 and later versions
            if (
                (
                    version[0] > 2017 or (version[0] == 2017 and version[1] >= 4)
                )  # 2017.4
                or (
                    (version[0] == 2017 and version[1] == 3 and version[2] == 1)
                    and self.build_type.IsPatch
                )  # fixed after 2017.3.1px
                or (
                    (version[0] == 2017 and version[1] == 3)
                    and self.m_MeshCompression == 0
                )
            ):  # 2017.3.xfx with no compression
                self.m_IndexFormat = reader.read_int()

            m_IndexBuffer_size = reader.read_int()
            if self.m_Use16BitIndices:
                self.m_IndexBuffer = [
                    reader.read_u_short() for _ in range(m_IndexBuffer_size // 2)
                ]
                reader.align_stream()
            else:
                self.m_IndexBuffer = reader.read_u_int_array(m_IndexBuffer_size // 4)

        if version[0] < 3 or (version[0] == 3 and version[1] < 5):  # 3.4.2 and earlier
            m_VertexCount = reader.read_int()
            self.m_Vertices = reader.read_float_array(m_VertexCount * 3)  # Vector3

            m_SkinSize = reader.read_int()
            self.m_Skin = [BoneWeights4(reader) for _ in range(m_SkinSize)]

            self.m_BindPose = reader.read_matrix_array()
            self.m_UV0 = reader.read_float_array(reader.read_int() * 2)  # Vector2
            self.m_UV1 = reader.read_float_array(reader.read_int() * 2)  # Vector2

            if version[0] == 2 and version[1] <= 5:  # 2.5 and down
                m_TangentSpace_size = reader.read_int()
                self.m_Normals = [0] * (m_TangentSpace_size * 3)
                self.m_Tangets = [0] * (m_TangentSpace_size * 4)
                for v in range(m_TangentSpace_size):
                    self.m_Normals[v * 3] = reader.read_float()
                    self.m_Normals[v * 3 + 1] = reader.read_float()
                    self.m_Normals[v * 3 + 2] = reader.read_float()
                    self.m_Tangents[v * 3] = reader.read_float()
                    self.m_Tangents[v * 3 + 1] = reader.read_float()
                    self.m_Tangents[v * 3 + 2] = reader.read_float()
                    self.m_Tangents[v * 3 + 3] = reader.read_float()  # handedness
            else:  # 2.6.0 and later
                self.m_Tangents = reader.read_float_array(
                    reader.read_int() * 4
                )  # Vector4
                self.m_Normals = reader.read_float_array(
                    reader.read_int() * 3
                )  # Vector3
        else:
            if version[0] < 2018 or (
                version[0] == 2018 and version[1] < 2
            ):  # 2018.2 down
                m_SkinSize = reader.read_int()
                self.m_Skin = [BoneWeights4(reader) for _ in range(m_SkinSize)]

            if version[0] == 3 or (version[0] == 4 and version[1] <= 2):  # 4.2 and down
                self.m_BindPose = reader.read_matrix_array()

            self.m_VertexData = VertexData(reader)

        if version[0] > 2 or (version[0] == 2 and version[1] >= 6):  # 2.6.0 and later
            self.m_CompressedMesh = CompressedMesh(reader)

        self.m_LocalAABB = AABB(reader)

        if version[0] < 3 or (version[0] == 3 and version[1] <= 4):  # 3.4.2 and earlier
            m_Colors_size = reader.read_int()
            self.mColors = [
                reader.read_byte()  # (float)reader.read_byte() / 0xFF
                for _ in range(m_Colors_size * 4)
            ]

            m_CollisionTriangles_size = reader.read_int()
            reader.Position += m_CollisionTriangles_size * 4  # UInt32 indices
            m_CollisionVertexCount = reader.read_int()

        self.m_MeshUsageFlags = reader.read_int()

        if version[0] >= 5:  # 5.0 and up
            self.m_BakedConvexCollisionMesh = reader.read_bytes(reader.read_int())
            reader.align_stream()
            self.m_BakedTriangleCollisionMesh = reader.read_bytes(reader.read_int())
            reader.align_stream()

        if version[0] > 2018 or (
            version[0] == 2018 and version[1] >= 2
        ):  # 2018.2 and up
            self.m_MeshMetrics = [reader.read_float(), reader.read_float()]

        if version[0] > 2018 or (
            version[0] == 2018 and version[1] >= 3
        ):  # 2018.3 and up
            reader.align_stream()
            self.m_StreamData = StreamingInfo(reader)

        # try:
        #    self.ProcessData()
        # except:
        #    pass

    def ProcessData(self):
        if self.m_StreamData and self.m_StreamData.path:
            if self.m_VertexData.m_VertexCount > 0:  #
                self.m_VertexData.m_DataSize = get_resource_data(
                    self.m_StreamData.path,
                    self.assets_file,
                    self.m_StreamData.offset,
                    self.m_StreamData.size,
                )
        # Fix channel after 2018.3
        version = self.version
        if version[0] > 2018 or (version[0] == 2018 and version[1] >= 3):  #
            self.m_VertexData.FixChannel()
        if version[0] > 3 or (version[0] == 3 and version[1] >= 5):  # 3.5 and up
            self.ReadVertexData()

        if version[0] > 2 or (version[0] == 2 and version[1] >= 6):  # 2.6.0 and later
            self.DecompressCompressedMesh()

        self.BuildFaces()

    def ReadVertexData(self):
        m_VertexCount = self.m_VertexData.m_VertexCount
        version = self.version
        for chn, m_Channel in enumerate(self.m_VertexData.m_Channels):
            if m_Channel.dimension > 0:
                m_Stream = self.m_VertexData.m_Streams[m_Channel.stream]
                channelMask = bytearray(m_Stream.channelMask)
                if channelMask[chn]:
                    if version[0] < 2018 and chn == 2 and m_Channel.format == 2:  #
                        m_Channel.dimension = 4

                    componentByteSize = MeshHelper.GetChannelFormatSize(
                        m_Channel.format
                    )
                    componentBytes = bytearray(
                        m_VertexCount * m_Channel.dimension * componentByteSize
                    )
                    for v in range(m_VertexCount):
                        vertexOffset = (
                            m_Stream.offset + m_Channel.offset + m_Stream.stride * v
                        )
                        for d in range(m_Channel.dimension):
                            componentOffset = vertexOffset + componentByteSize * d
                            dstOffset = componentByteSize * (
                                v * m_Channel.dimension + d
                            )
                            componentBytes[
                                dstOffset : dstOffset + componentByteSize
                            ] += self.m_VertexData.m_DataSize[
                                componentOffset : componentOffset + componentByteSize
                            ]

                    if (
                        self.reader.endian == ">" and componentByteSize > 1
                    ):  # if big endian, swap bytes
                        for i in range(0, len(componentBytes), componentByteSize):
                            componentBytes[i : i + componentByteSize] = componentBytes[
                                i : i + componentByteSize
                            ][::-1]

                    componentsIntArray = []
                    componentsFloatArray = []
                    if m_Channel.format == 10 or m_Channel.format == 11:  #
                        componentsIntArray = MeshHelper.BytesToIntArray(componentBytes)
                    else:
                        componentsFloatArray = MeshHelper.BytesToFloatArray(
                            componentBytes, componentByteSize
                        )

                    if version[0] >= 2018:  #
                        if chn == 0:  # kShaderChannelVertex
                            self.m_Vertices = componentsFloatArray
                        elif chn == 1:  # kShaderChannelNormal
                            self.m_Normals = componentsFloatArray
                        elif chn == 2:  # kShaderChannelTangent
                            self.m_Tangents = componentsFloatArray
                        elif chn == 3:  # kShaderChannelColor
                            self.m_Colors = componentsFloatArray
                        elif chn == 4:  # kShaderChannelTexCoord0
                            self.m_UV0 = componentsFloatArray
                        elif chn == 5:  # kShaderChannelTexCoord1
                            self.m_UV1 = componentsFloatArray
                        elif chn == 6:  # kShaderChannelTexCoord2
                            self.m_UV2 = componentsFloatArray
                        elif chn == 7:  # kShaderChannelTexCoord3
                            self.m_UV3 = componentsFloatArray
                        # kShaderChannelTexCoord4 8
                        # kShaderChannelTexCoord5 9
                        # kShaderChannelTexCoord6 10
                        # kShaderChannelTexCoord7 11
                        # 2018.2 and up
                        elif chn == 12:  # kShaderChannelBlendWeight
                            if getattr(self, "m_Skin", None):
                                self.InitMSkin()
                            for i in range(m_VertexCount):
                                for j in range(m_Channel.dimension):
                                    self.m_Skin[i].weight[j] = componentsFloatArray[
                                        i * m_Channel.dimension + j
                                    ]
                        elif chn == 13:  # kShaderChannelBlendIndices
                            if getattr(self, "m_Skin", None):
                                self.InitMSkin()
                            for i in range(m_VertexCount):
                                for j in range(m_Channel.dimension):
                                    self.m_Skin[i].boneIndex[j] = componentsIntArray[
                                        i * m_Channel.dimension + j
                                    ]

                    else:
                        if chn == 0:  # kShaderChannelVertex
                            self.m_Vertices = componentsFloatArray
                        elif chn == 1:  # kShaderChannelNormal
                            self.m_Normals = componentsFloatArray
                        elif chn == 2:  # kShaderChannelColor
                            self.m_Colors = componentsFloatArray
                        elif chn == 3:  # kShaderChannelTexCoord0
                            self.m_UV0 = componentsFloatArray
                        elif chn == 4:  # kShaderChannelTexCoord1
                            self.m_UV1 = componentsFloatArray
                        elif chn == 5:
                            if version[0] >= 5:  # kShaderChannelTexCoord2
                                self.m_UV2 = componentsFloatArray
                            else:  # kShaderChannelTangent
                                self.m_Tangents = componentsFloatArray
                        elif chn == 6:  # kShaderChannelTexCoord3
                            self.m_UV3 = componentsFloatArray
                        elif chn == 7:  # kShaderChannelTangent
                            self.m_Tangents = componentsFloatArray

    def DecompressCompressedMesh(self):
        # Vertex
        version = self.version
        m_CompressedMesh = self.m_CompressedMesh
        if m_CompressedMesh.m_Vertices.m_NumItems > 0:
            m_VertexCount = m_CompressedMesh.m_Vertices.m_NumItems / 3
            m_Vertices = m_CompressedMesh.m_Vertices.UnpackFloats(3, 4)
        # UV
        if m_CompressedMesh.m_UV.m_NumItems > 0:  #
            m_UV0 = m_CompressedMesh.m_UV.UnpackFloats(2, 4, 0, m_VertexCount)
            if m_CompressedMesh.m_UV.m_NumItems >= m_VertexCount * 4:  #
                m_UV1 = m_CompressedMesh.m_UV.UnpackFloats(
                    2, 4, m_VertexCount * 2, m_VertexCount
                )
            if m_CompressedMesh.m_UV.m_NumItems >= m_VertexCount * 6:  #
                m_UV2 = m_CompressedMesh.m_UV.UnpackFloats(
                    2, 4, m_VertexCount * 4, m_VertexCount
                )
            if m_CompressedMesh.m_UV.m_NumItems >= m_VertexCount * 8:  #
                m_UV3 = m_CompressedMesh.m_UV.UnpackFloats(
                    2, 4, m_VertexCount * 6, m_VertexCount
                )
        # BindPose
        if version[0] < 5:  #
            if m_CompressedMesh.m_BindPoses.m_NumItems > 0:  #
                m_BindPoses_Unpacked = m_CompressedMesh.m_BindPoses.UnpackFloats(
                    16, 4 * 16
                )
                self.m_BindPose = [
                    Matrix4x4(m_BindPoses_Unpacked[i : i + 16])
                    for i in range(0, m_CompressedMesh.m_BindPoses.m_NumItems, 16)
                ]

        # Normal
        if m_CompressedMesh.m_Normals.m_NumItems > 0:
            normalData = m_CompressedMesh.m_Normals.UnpackFloats(2, 4 * 2)
            signs = m_CompressedMesh.m_NormalSigns.UnpackInts()
            self.m_Normals = []  # float[m_CompressedMesh.m_Normals.m_NumItems / 2 * 3]
            for i in range(0, math.ceil(m_CompressedMesh.m_Normals.m_NumItems / 2)):
                x = normalData[i + 0]
                y = normalData[i + 1]
                zsqr = 1 - x * x - y * y
                if zsqr >= 0:
                    z = math.sqrt(zsqr)
                else:
                    z = 0
                    normal = Vector3(x, y, z)
                    normal.normalize()
                    x = normal.X
                    y = normal.Y
                    z = normal.Z
                if signs[i] == 0:
                    z = -z
                self.m_Normals.extend([x, y, z])
        # Tangent
        if m_CompressedMesh.m_Tangents.m_NumItems > 0:
            tangentData = m_CompressedMesh.m_Tangents.UnpackFloats(2, 4 * 2)
            signs = m_CompressedMesh.m_TangentSigns.UnpackInts()
            self.m_Tangents = (
                []
            )  # float[m_CompressedMesh.m_Tangents.m_NumItems / 2 * 4]
            for i in range(0, math.ceil(m_CompressedMesh.m_Tangents.m_NumItems / 2)):
                x = tangentData[i * 2 + 0]
                y = tangentData[i * 2 + 1]
                zsqr = 1 - x * x - y * y
                if zsqr >= 0:
                    z = math.sqrt(zsqr)
                else:
                    z = 0
                    vector3f = Vector3(x, y, z)
                    vector3f.Normalize()
                    x = vector3f.X
                    y = vector3f.Y
                    z = vector3f.Z
                if signs[i * 2 + 0] == 0:  #
                    z = -z
                w = 1.0 if signs[i * 2 + 1] > 0 else -1.0
                self.m_Tangents.extend([x, y, z, w])

        # FloatColor
        if version[0] >= 5:  #
            if m_CompressedMesh.m_FloatColors.m_NumItems > 0:  #
                self.m_Colors = m_CompressedMesh.m_FloatColors.UnpackFloats(1, 4)
        # Skin
        if m_CompressedMesh.m_Weights.m_NumItems > 0:
            weights = m_CompressedMesh.m_Weights.UnpackInts()
            boneIndices = m_CompressedMesh.m_BoneIndices.UnpackInts()

            self.InitMSkin()

            bonePos = 0
            boneIndexPos = 0
            j = 0
            sum = 0

            for i in range(m_CompressedMesh.m_Weights.m_NumItems):
                # read bone index and weight.
                self.m_Skin[bonePos].weight[j] = weights[i] / 31.0
                self.m_Skin[bonePos].boneIndex[j] = boneIndices[boneIndexPos]
                boneIndexPos += 1
                j += 1
                sum += weights[i]

                # the weights add up to one. fill the rest for this vertex with zero, and continue with next one.
                if sum >= 31:  #
                    while j < 4:
                        self.m_Skin[bonePos].weight[j] = 0
                        self.m_Skin[bonePos].boneIndex[j] = 0
                        bonePos += 1
                        j += 1
                    j = 0
                    sum = 0
                # we read three weights, but they don't add up to one. calculate the fourth one, and read
                # missing bone index. continue with next vertex.
                elif j == 3:  #
                    self.m_Skin[bonePos].weight[j] = (31 - sum) / 31.0
                    self.m_Skin[bonePos].boneIndex[j] = boneIndices[boneIndexPos]
                    boneIndexPos += 1
                    bonePos += 1
                    j = 0
                    sum = 0

        # IndexBuffer
        if m_CompressedMesh.m_Triangles.m_NumItems > 0:  #
            self.m_IndexBuffer = m_CompressedMesh.m_Triangles.UnpackUInts()
        # Color
        if (
            hasattr(m_CompressedMesh, "m_Colors")
            and m_CompressedMesh.m_Colors.m_NumItems > 0
        ):
            m_CompressedMesh.m_Colors.m_NumItems *= 4
            m_CompressedMesh.m_Colors.m_BitSize /= 4
            tempColors = m_CompressedMesh.m_Colors.UnpackInts()
            self.m_Colors = [color / 255 for color in tempColors]

    def BuildFaces(self):
        for m_SubMesh in self.m_SubMeshes:
            firstIndex = m_SubMesh.firstByte // 2
            if not self.m_Use16BitIndices:
                firstIndex /= 2

            if m_SubMesh.topology == 0:
                for i in range(0, m_SubMesh.indexCount, 3):
                    self.m_Indices.append(self.m_IndexBuffer[firstIndex + i])
                    self.m_Indices.append(self.m_IndexBuffer[firstIndex + i + 1])
                    self.m_Indices.append(self.m_IndexBuffer[firstIndex + i + 2])
            else:
                j = 0
                for i in range(0, m_SubMesh.indexCount - 2):
                    fa = self.m_IndexBuffer[firstIndex + i]
                    fb = self.m_IndexBuffer[firstIndex + i + 1]
                    fc = self.m_IndexBuffer[firstIndex + i + 2]

                    if (fa != fb) and (fa != fc) and (fc != fb):
                        self.m_Indices.append(fa)
                        self.m_Indices.extend([fc, fb] if i % 2 else [fb, fc])
                        j += 1
                # fix indexCount
                m_SubMesh.indexCount = j * 3

    def InitMSkin(self):
        self.m_Skin = [BoneWeights4() for _ in range(self.m_VertexCount)]


class MeshHelper:
    def GetChannelFormatSize(format):
        if format == 0:  # kChannelFormatFloat
            return 4
        elif format == 1:  # kChannelFormatFloat16
            return 2
        elif format == 2:  # kChannelFormatColor
            return 1
        elif format == 3:  # kChannelFormatByte
            return 1
        elif format == 4:  # kChannelFormatUInt32
            return 4
        elif format == 10:  # kChannelFormatInt32
            return 4
        elif format == 11:  # kChannelFormatInt32
            return 4
        else:
            return 0

    def BytesToFloatArray(inputBytes, size):
        result = []
        for i in range(math.ceil(len(inputBytes) / size)):
            value = 0
            if size == 1:
                value = inputBytes[i] / 255.0
            elif size == 2:
                value = float.from_bytes(
                    inputBytes[i * 2 : i * 2 + 2], byteorder="big", signed=True
                )  # return Half.ToHalf((ushort)BitConverter.ToInt16(value, startIndex));
            elif size == 4:
                value = float.from_bytes(
                    inputBytes[i * 4 : i * 4 + 4], byteorder="big", signed=True
                )
            result[i] = value
        return result

    def BytesToIntArray(inputBytes):
        return [
            int.from_bytes(inputBytes[i : i + 4], byteorder="big", signed=True)
            for i in range(0, len(inputBytes), 4)
        ]



import math
from enum import IntEnum

from .NamedObject import NamedObject
from .PPtr import PPtr
from ..enums import ClassIDType
from ..math import Quaternion, Vector3
from ..streams import EndianBinaryReader


def uint(num):
    if num < 0 or num > 4294967295:
        return num % 4294967296
    return num


class Keyframe:
    def __init__(self, reader, readerFunc):
        self.time = reader.read_float()
        self.value = readerFunc()
        self.inSlope = readerFunc()
        self.outSlope = readerFunc()
        if reader.version[0] >= 2018:  # 2018 and up
            self.weightedMode = reader.read_int()
            self.inWeight = readerFunc()
            self.outWeight = readerFunc()


class AnimationCurve:
    def __init__(self, reader, readerFunc):
        version = reader.version
        numCurves = reader.read_int()
        self.m_Curve = [Keyframe(reader, readerFunc) for _ in range(numCurves)]

        self.m_PreInfinity = reader.read_int()
        self.m_PostInfinity = reader.read_int()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 3):  # 5.3 and up
            self.m_RotationOrder = reader.read_int()


class QuaternionCurve:
    def __init__(self, reader):
        self.curve = AnimationCurve(reader, reader.read_quaternion)  # <Quaternion>
        self.path = reader.read_aligned_string()


class PackedFloatVector:
    def __init__(self, reader):
        self.m_NumItems = reader.read_u_int()
        self.m_Range = reader.read_float()
        self.m_Start = reader.read_float()
        numData = reader.read_int()
        self.m_Data = reader.read_bytes(numData)
        reader.align_stream()
        self.m_BitSize = reader.read_byte()
        reader.align_stream()

    def UnpackFloats(
        self,
        itemCountInChunk: int,
        chunkStride: int,
        start: int = 0,
        numChunks: int = -1,
    ):
        bitPos: int = self.m_BitSize * start
        indexPos: int = bitPos // 8
        bitPos %= 8

        scale: float = 1.0 / self.m_Range
        if numChunks == -1:
            numChunks = self.m_NumItems // itemCountInChunk
        end = chunkStride * numChunks / 4
        data = []
        for index in (0, end, chunkStride // 4):
            for i in range(itemCountInChunk):
                x = 0  # uint
                bits = 0
                while bits < self.m_BitSize:
                    x |= uint(
                        (self.m_Data[indexPos] >> bitPos) << bits
                    )  # (uint)((m_Data[indexPos] >> bitPos) << bits)
                    num = min(self.m_BitSize - bits, 8 - bitPos)
                    bitPos += num
                    bits += num
                    if bitPos == 8:  #
                        indexPos += 1
                        bitPos = 0

                x &= uint((1 << self.m_BitSize) - 1)  # (uint)(1 << m_BitSize) - 1u
                data.append(x / (scale * ((1 << self.m_BitSize) - 1)) + self.m_Start)

        return data


class PackedIntVector:
    def __init__(self, reader):
        self.m_NumItems = reader.read_u_int()
        numData = reader.read_int()
        self.m_Data = reader.read_bytes(numData)
        reader.align_stream()

        self.m_BitSize = reader.read_byte()
        reader.align_stream()

    def UnpackInts(self):
        data = []
        indexPos = 0
        bitPos = 0
        m_BitSize = self.m_BitSize

        for i in range(self.m_NumItems):
            bits = 0
            entry = 0
            while bits << m_BitSize:
                entry |= (self.m_Data[indexPos] >> bitPos) << bits
                num = min(m_BitSize - bits, 8 - bitPos)
                bitPos += num
                bits += num
                if bitPos == 8:  #
                    indexPos += 1
                    bitPos = 0
            data.append(entry & (1 << m_BitSize) - 1)
        return data


class PackedQuatVector:
    def __init__(self, reader):
        self.m_NumItems = reader.read_u_int()
        numData = reader.read_int()
        self.m_Data = reader.read_bytes(numData)
        reader.align_stream()

    def UnpackQuats(self):
        m_Data = self.m_Data
        data = []
        indexPos = 0
        bitPos = 0

        for i in range(self.m_NumItems):
            flags = 0
            bits = 0
            while bits < 3:
                flags |= (m_Data[indexPos] >> bitPos) << bits  # unit
                num = min(3 - bits, 8 - bitPos)
                bitPos += num
                bits += num
                if bitPos == 8:  #
                    indexPos += 1
                    bitPos = 0
            flags &= 7

            q = Quaternion()
            sum = 0
            for j in range(4):
                if (flags & 3) != j:  #
                    bitSize = 9 if ((flags & 3) + 1) % 4 == j else 10
                    x = 0

                    bits = 0
                    while bits < bitSize:
                        x |= (m_Data[indexPos] >> bitPos) << bits  # uint
                        num = min(bitSize - bits, 8 - bitPos)
                        bitPos += num
                        bits += num
                        if bitPos == 8:  #
                            indexPos += 1
                            bitPos = 0
                    x &= (1 << bitSize) - 1  # unit
                    q[j] = x / (0.5 * ((1 << bitSize) - 1)) - 1
                    sum += q[j] * q[j]

            lastComponent = flags & 3  # int
            q[lastComponent] = math.sqrt(1 - sum)  # float
            if (flags & 4) != 0:  # 0u
                q[lastComponent] = -q[lastComponent]
            data.append(q)

        return data


class CompressedAnimationCurve:
    def __init__(self, reader):
        self.m_Path = reader.read_aligned_string()
        self.m_Times = PackedIntVector(reader)
        self.m_Values = PackedQuatVector(reader)
        self.m_Slopes = PackedFloatVector(reader)
        self.m_PreInfinity = reader.read_int()
        self.m_PostInfinity = reader.read_int()


class Vector3Curve:
    def __init__(self, reader):
        self.curve = AnimationCurve(reader, reader.read_vector3)  # Vector3
        self.path = reader.read_aligned_string()


class FloatCurve:
    def __init__(self, reader):
        self.curve = AnimationCurve(reader, reader.read_float)  # Float
        self.attribute = reader.read_aligned_string()
        self.path = reader.read_aligned_string()
        self.classID = reader.read_int()
        self.script = PPtr(reader)  # MonoScript


class PPtrKeyframe:
    def __init__(self, reader):
        self.time = reader.read_float()
        self.value = PPtr(reader)  # Object


class PPtrCurve:
    def __init__(self, reader):
        numCurves = reader.read_int()
        self.curve = [PPtrKeyframe(reader) for _ in range(numCurves)]

        self.attribute = reader.read_aligned_string()
        self.path = reader.read_aligned_string()
        self.classID = reader.read_int()
        self.script = PPtr(reader)  # MonoScript


class AABB:
    def __init__(self, reader):
        self.m_Center = reader.read_vector3()
        self.m_Extent = reader.read_vector3()


class xform:
    def __init__(self, reader):
        version = reader.version
        self.t = (
            reader.read_vector3()
            if version[0] > 5 or (version[0] == 5 and version[1] >= 4)
            else Vector3(reader.read_vector4())
        )  # 5.4 and up
        self.q = reader.read_quaternion()
        self.s = (
            reader.read_vector3()
            if version[0] > 5 or (version[0] == 5 and version[1] >= 4)
            else Vector3(reader.read_vector4())
        )  # 5.4 and up


class HandPose:
    def __init__(self, reader):
        self.m_GrabX = xform(reader)
        self.m_DoFArray = reader.read_float_array()
        self.m_Override = reader.read_float()
        self.m_CloseOpen = reader.read_float()
        self.m_InOut = reader.read_float()
        self.m_Grab = reader.read_float()


class HumanGoal:
    def __init__(self, reader):
        version = reader.version
        self.m_X = xform(reader)
        self.m_WeightT = reader.read_float()
        self.m_WeightR = reader.read_float()
        if version[0] >= 5:  # 5.0 and up
            self.m_HintT = (
                reader.read_vector3()
                if version[0] > 5 or (version[0] == 5 and version[1] >= 4)
                else Vector3(reader.read_vector4())
            )  # 5.4 and up
            self.m_HintWeightT = reader.read_float()


class HumanPose:
    def __init__(self, reader):
        version = reader.version
        self.m_RootX = xform(reader)
        self.m_LookAtPosition = (
            reader.read_vector3()
            if version[0] > 5 or (version[0] == 5 and version[1] >= 4)
            else Vector3(reader.read_vector4())
        )  # 5.4 and up
        self.m_LookAtWeight = reader.read_vector4()

        numGoals = reader.read_int()
        self.m_GoalArray = [HumanGoal(reader) for _ in range(numGoals)]

        self.m_LeftHandPose = HandPose(reader)
        self.m_RightHandPose = HandPose(reader)

        self.m_DoFArray = reader.read_float_array()

        if version[0] > 5 or (version[0] == 5 and version[1] >= 2):  # 5.2 and up
            numTDof = reader.read_int()
            self.m_TDoFArray = [
                reader.read_vector3()
                if version[0] > 5 or (version[0] == 5 and version[1] >= 4)
                else Vector3(reader.read_vector4())  # 5.4 and up
                for _ in range(numTDof)
            ]


class StreamedCurveKey:
    def __init__(self, reader):
        self.index = reader.read_int()
        self.coeff = reader.read_float_array(4)

        self.outSlope = self.coeff[2]
        self.value = self.coeff[3]

    def CalculateNextInSlope(self, dx: float, rhs):
        """
        :param dx: float
        :param rhs: StreamedCurvedKey
        :return:
        """
        # Stepped
        if self.coeff[0] == 0 and self.coeff[1] == 0 and self.coeff[2] == 0:
            return float.PositiveInfinity

        dx = max(dx, 0.0001)
        dy = rhs.value - self.value
        length = 1.0 / (dx * dx)
        d1 = self.outSlope * dx
        d2 = dy + dy + dy - d1 - d1 - self.coeff[1] / length
        return d2 / dx


class StreamedFrame:
    def __init__(self, reader):
        self.time = reader.read_float()
        numKeys = reader.read_int()
        self.keyList = [StreamedCurveKey(reader) for _ in range(numKeys)]


class StreamedClip:
    def __init__(self, reader):
        self.data = reader.read_u_int_array()
        self.curveCount = reader.read_u_int()

    def ReadData(self):
        frameList = []
        buffer = self.data[0 : len(self.data) * 4]
        reader = EndianBinaryReader(buffer)
        while reader.Position < reader.Length:
            frameList.append(StreamedFrame(reader))

        for frameIndex in range(2, len(frameList) - 1):
            frame = frameList[frameIndex]
            for curveKey in frame.keyList:
                i = frameIndex - 1
                while i >= 0:
                    preFrame = frameList[i]
                    try:
                        preCurveKey = [
                            x for x in preFrame.keyList if x.index == curveKey.index
                        ][0]
                        curveKey.inSlope = preCurveKey.CalculateNextInSlope(
                            frame.time - preFrame.time, curveKey
                        )
                        break
                    except IndexError:
                        pass
                    i -= 1
        return frameList


class DenseClip:
    def __init__(self, reader):
        self.m_FrameCount = reader.read_int()
        self.m_CurveCount = reader.read_u_int()
        self.m_SampleRate = reader.read_float()
        self.m_BeginTime = reader.read_float()
        self.m_SampleArray = reader.read_float_array()


class ConstantClip:
    def __init__(self, reader):
        self.data = reader.read_float_array()


class ValueConstant:
    def __init__(self, reader):
        version = reader.version
        self.m_ID = reader.read_u_int()
        if version[0] < 5 or (version[0] == 5 and version[1] < 5):  # 5.5 down
            self.m_TypeID = reader.read_u_int()
        self.m_Type = reader.read_u_int()
        self.m_Index = reader.read_u_int()


class ValueArrayConstant:
    def __init__(self, reader):
        numVals = reader.read_int()
        self.m_ValueArray = [ValueConstant(reader) for _ in range(numVals)]


class Clip:
    def __init__(self, reader):
        version = reader.version
        self.m_StreamedClip = StreamedClip(reader)
        self.m_DenseClip = DenseClip(reader)
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_ConstantClip = ConstantClip(reader)
        if version[0] < 2018 or (version[0] == 2018 and version[1] < 3):  # 2018.3 down
            self.m_Binding = ValueArrayConstant(reader)


class ValueDelta:
    def __init__(self, reader):
        self.m_Start = reader.read_float()
        self.m_Stop = reader.read_float()


class ClipMuscleConstant:
    def __init__(self, reader):
        version = reader.version
        self.m_DeltaPose = HumanPose(reader)
        self.m_StartX = xform(reader)
        if version[0] > 5 or (version[0] == 5 and version[1] >= 5):  # 5.5 and up
            self.m_StopX = xform(reader)
        self.m_LeftFootStartX = xform(reader)
        self.m_RightFootStartX = xform(reader)
        if version[0] < 5:  # 5.0 down
            self.m_MotionStartX = xform(reader)
            self.m_MotionStopX = xform(reader)
        self.m_AverageSpeed = (
            reader.read_vector3()
            if version[0] > 5 or (version[0] == 5 and version[1] >= 4)
            else Vector3(reader.read_vector4())
        )  # 5.4 and up
        self.m_Clip = Clip(reader)
        self.m_StartTime = reader.read_float()
        self.m_StopTime = reader.read_float()
        self.m_OrientationOffsetY = reader.read_float()
        self.m_Level = reader.read_float()
        self.m_CycleOffset = reader.read_float()
        self.m_AverageAngularSpeed = reader.read_float()

        self.m_IndexArray = reader.read_int_array()
        if version[0] < 4 or (version[0] == 4 and version[1] < 3):  # 4.3 down
            self.m_AdditionalCurveIndexArray = reader.read_int_array()
        numDeltas = reader.read_int()
        self.m_ValueArrayDelta = [ValueDelta(reader) for _ in range(numDeltas)]
        if version[0] > 5 or (version[0] == 5 and version[1] >= 3):  # 5.3 and up
            self.m_ValueArrayReferencePose = reader.read_float_array()

        self.m_Mirror = reader.read_boolean()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_LoopTime = reader.read_boolean()
        self.m_LoopBlend = reader.read_boolean()
        self.m_LoopBlendOrientation = reader.read_boolean()
        self.m_LoopBlendPositionY = reader.read_boolean()
        self.m_LoopBlendPositionXZ = reader.read_boolean()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 5):  # 5.5 and up
            self.m_StartAtOrigin = reader.read_boolean()
        self.m_KeepOriginalOrientation = reader.read_boolean()
        self.m_KeepOriginalPositionY = reader.read_boolean()
        self.m_KeepOriginalPositionXZ = reader.read_boolean()
        self.m_HeightFromFeet = reader.read_boolean()
        reader.align_stream()


class GenericBinding:
    def __init__(self, reader):
        version = reader.version
        self.path = reader.read_u_int()
        self.attribute = reader.read_u_int()
        self.script = PPtr(reader)  # Object
        if version[0] > 5 or (version[0] == 5 and version[1] >= 6):  # 5.6 and up
            self.typeID = ClassIDType(reader.read_int())
        else:
            self.typeID = ClassIDType(reader.read_u_short())
        self.customType = reader.read_byte()
        self.isPPtrCurve = reader.read_byte()
        reader.align_stream()


class AnimationClipBindingConstant:
    def __init__(self, reader):
        numBindings = reader.read_int()
        self.genericBindings = [GenericBinding(reader) for _ in range(numBindings)]

        numMappings = reader.read_int()
        self.pptrCurveMapping = [PPtr(reader) for _ in range(numMappings)]  # Object

        def FindBinding(self, index):
            curves = 0
            for b in self.genericBindings:
                if b.typeID == ClassIDType.Transform:  #
                    switch = b.attribute

                    if switch in [1, 3, 4]:
                        # case 1: #kBindTransformPosition
                        # case 3: #kBindTransformScale
                        # case 4: #kBindTransformEuler
                        curves += 3
                    elif switch == 2:  # kBindTransformRotation
                        curves += 4
                    else:
                        curves += 1
                else:
                    curves += 1
                if curves > index:
                    return b
            return None


class AnimationType(IntEnum):
    kLegacy = (1,)
    kGeneric = (2,)
    kHumanoid = 3


class AnimationClip(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = reader.version
        if version[0] >= 5:  # 5.0 and up
            self.m_Legacy = reader.read_boolean()
        elif version[0] >= 4:  # 4.0 and up
            self.m_AnimationType = AnimationType(reader.read_int())
            if self.m_AnimationType == AnimationType.kLegacy:  #
                self.m_Legacy = True
        else:
            self.m_Legacy = True

        self.m_Compressed = reader.read_boolean()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_UseHighQualityCurve = reader.read_boolean()
        reader.align_stream()
        numRCurves = reader.read_int()
        self.m_RotationCurves = [QuaternionCurve(reader) for _ in range(numRCurves)]

        numCRCurves = reader.read_int()
        self.m_CompressedRotationCurves = [
            CompressedAnimationCurve(reader) for _ in range(numCRCurves)
        ]

        if version[0] > 5 or (version[0] == 5 and version[1] >= 3):  # 5.3 and up
            numEulerCurves = reader.read_int()
            self.m_EulerCurves = [Vector3Curve(reader) for _ in range(numEulerCurves)]

        numPCurves = reader.read_int()
        self.m_PositionCurves = [Vector3Curve(reader) for _ in range(numPCurves)]

        numSCurves = reader.read_int()
        self.m_ScaleCurves = [Vector3Curve(reader) for _ in range(numSCurves)]

        numFCurves = reader.read_int()
        self.m_FloatCurves = [FloatCurve(reader) for _ in range(numFCurves)]
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            numPtrCurves = reader.read_int()
            self.m_PPtrCurves = [PPtrCurve(reader) for _ in range(numPtrCurves)]

        self.m_SampleRate = reader.read_float()
        self.m_WrapMode = reader.read_int()
        if version[0] > 3 or (version[0] == 3 and version[1] >= 4):  # 3.4 and up
            self.m_Bounds = AABB(reader)
        if version[0] >= 4:  # 4.0 and up
            self.m_MuscleClipSize = reader.read_u_int()
            self.m_MuscleClip = ClipMuscleConstant(reader)
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_ClipBindingConstant = AnimationClipBindingConstant(reader)


# m_HasGenericRootTransform 2018.3
# m_HasMotionFloatCurves 2018.3
# numEvents = reader.read_int()
# self.m_Events = [
# 	AnimationEvent(reader)
# 	for _ in range(numEvents)
# ]



COMMON_STRING = {
    0: "AABB",
    5: "AnimationClip",
    19: "AnimationCurve",
    34: "AnimationState",
    49: "Array",
    55: "Base",
    60: "BitField",
    69: "bitset",
    76: "bool",
    81: "char",
    86: "ColorRGBA",
    96: "Component",
    106: "data",
    111: "deque",
    117: "double",
    124: "dynamic_array",
    138: "FastPropertyName",
    155: "first",
    161: "float",
    167: "Font",
    172: "GameObject",
    183: "Generic Mono",
    196: "GradientNEW",
    208: "GUID",
    213: "GUIStyle",
    222: "int",
    226: "list",
    231: "long long",
    241: "map",
    245: "Matrix4x4f",
    256: "MdFour",
    263: "MonoBehaviour",
    277: "MonoScript",
    288: "byte_size",
    299: "m_Curve",
    307: "m_EditorClassIdentifier",
    331: "m_EditorHideFlags",
    349: "m_Enabled",
    359: "m_ExtensionPtr",
    374: "m_GameObject",
    387: "index",
    395: "is_array",
    405: "m_IsStatic",
    416: "meta_flag",
    427: "name",
    434: "_object_hide_flags",
    452: "m_PrefabInternal",
    469: "m_PrefabParentObject",
    490: "m_Script",
    499: "m_StaticEditorFlags",
    519: "type",
    526: "version",
    536: "Object",
    543: "pair",
    548: "PPtr<Component>",
    564: "PPtr<GameObject>",
    581: "PPtr<Material>",
    596: "PPtr<MonoBehaviour>",
    616: "PPtr<MonoScript>",
    633: "PPtr<Object>",
    646: "PPtr<Prefab>",
    659: "PPtr<Sprite>",
    672: "PPtr<TextAsset>",
    688: "PPtr<Texture>",
    702: "PPtr<Texture2D>",
    718: "PPtr<Transform>",
    734: "Prefab",
    741: "Quaternionf",
    753: "Rectf",
    759: "RectInt",
    767: "RectOffset",
    778: "second",
    785: "set",
    789: "short",
    795: "size",
    800: "SInt16",
    807: "SInt32",
    814: "SInt64",
    821: "SInt8",
    827: "staticvector",
    840: "string",
    847: "TextAsset",
    857: "TextMesh",
    866: "Texture",
    874: "Texture2D",
    884: "Transform",
    894: "TypelessData",
    907: "UInt16",
    914: "UInt32",
    921: "UInt64",
    928: "UInt8",
    934: "unsigned int",
    947: "unsigned long long",
    966: "unsigned short",
    981: "vector",
    988: "Vector2f",
    997: "Vector3f",
    1006: "Vector4f",
    1015: "m_ScriptingClassIdentifier",
    1042: "Gradient",
    1051: "Type*",
    1057: "int2_storage",
    1070: "int3_storage",
    1083: "BoundsInt",
    1093: "m_CorrespondingSourceObject",
    1121: "m_PrefabInstance",
    1138: "m_PrefabAsset",
}



from .Behaviour import Behaviour
from .PPtr import PPtr


class Animator(Behaviour):
    def __init__(self, reader):
        super().__init__(reader=reader)

        self.avatar = PPtr(reader)  # Avatar
        self.controller = PPtr(reader)  # RuntimeAnimatorController
        self.culling_mode = reader.read_int()

        if self.version[0] > 4 or (self.version[0] == 4 and self.version[1] > 4):
            self.update_mode = reader.read_int()

        self.m_ApplyRootMotion = reader.read_boolean()
        if self.version[0] == 4 and self.version[1] >= 5:  # 4.5 and up - 5.0 down
            reader.align_stream()

        if self.version[0] >= 5:  # 5.0 and up
            self.linear_velocity_blending = reader.read_boolean()
            reader.align_stream()

        if self.version[0] < 4 or (
            self.version[0] == 4 and self.version[1] < 5
        ):  # 4.5 down
            self.animate_physics = reader.read_boolean()

        if self.version[0] > 4 or (
            self.version[0] == 4 and self.version[1] >= 3
        ):  # 4.3 and up
            self.has_transform_hierarchy = reader.read_boolean()

        if self.version[0] > 4 or (
            self.version[0] == 4 and self.version[1] >= 5
        ):  # 4.5 and up
            self.allow_constant_clip_sampling_optimization = reader.read_boolean()

        if 4 < self.version[0] < 2018:  # 5.0 and up - 2018 down
            reader.align_stream()

        if self.version[0] >= 2018:  # 2018 and up
            self.m_KeepAnimatorControllerStateOnDisable = reader.read_boolean()
            reader.align_stream()



from .Object import Object
from .PPtr import PPtr, save_ptr
from ..enums import BuildTarget
from ..streams import EndianBinaryWriter, EndianBinaryReader


class EditorExtension(Object):
    def __init__(self, reader: EndianBinaryReader):
        super().__init__(reader=reader)
        if self.platform == BuildTarget.NoTarget:
            self.prefab_parent_object = PPtr(reader)
            self.prefab_internal = PPtr(reader)

    def save(self, writer: EndianBinaryWriter):
        super().save(writer, intern_call=True)
        if self.platform == BuildTarget.NoTarget:
            save_ptr(self.prefab_parent_object)
            save_ptr(self.prefab_internal)



import os

from .ImportHelper import find_all_files


def get_resource_data(*args):
    """
    Input:
    Option 1:
        path - file path
        assets_file - SerializedFile
        offset -
        size -
    Option 2:
        reader - EndianBinaryReader
        offset -
        size -
    """
    if len(args) == 4:
        need_search = True
        path = args[0]
        assets_file = args[1]
    elif len(args) == 3:
        need_search = False
        reader = args[0]
    offset = args[-2]
    size = args[-1]

    if need_search:
        resource_file_name = path

        reader = assets_file.environment.resources.get(resource_file_name)
        if not reader:
            reader = assets_file.environment.resources.get(
                os.path.basename(resource_file_name)
            )
        if reader:
            reader.Position = offset
            return reader.read_bytes(size)

        if not assets_file.environment.path:
            return
        current_directory = os.path.dirname(assets_file.environment.path)
        resource_file_path = os.path.join(current_directory, resource_file_name)
        if not os.path.isfile(resource_file_path):
            find_files = find_all_files(current_directory, resource_file_name)
            if find_files:
                resource_file_path = find_files[0]

        if os.path.isfile(resource_file_path):
            with open(resource_file_path, "rb") as f:
                f.seek(offset)
                return f.read(size)
        else:
            raise FileNotFoundError(
                f"Can't find the resource file {resource_file_name}"
            )

    else:
        reader.Position = offset
        return reader.read_bytes(size)



class Quaternion:
    X: float
    Y: float
    Z: float
    W: float

    def __init__(self, x: float, y: float, z: float, w: float):
        self.X = x
        self.Y = y
        self.Z = z
        self.W = w


"""
using System;
using System.Runtime.InteropServices;

namespace AssetStudio
{
    [StructLayout(LayoutKind.Sequential, Pack = 4)]
    public struct Quaternion : IEquatable<Quaternion>
    {
        public float X;
        public float Y;
        public float Z;
        public float W;

        public Quaternion(float x, float y, float z, float w)
        {
            X = x;
            Y = y;
            Z = z;
            W = w;
        }

        public float this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return X;
                    case 1: return Y;
                    case 2: return Z;
                    case 3: return W;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Quaternion index!");
                }
            }

            set
            {
                switch (index)
                {
                    case 0: X = value; break;
                    case 1: Y = value; break;
                    case 2: Z = value; break;
                    case 3: W = value; break;
                    default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Quaternion index!");
                }
            }
        }

        public override int GetHashCode()
        {
            return X.GetHashCode() ^ (Y.GetHashCode() << 2) ^ (Z.GetHashCode() >> 2) ^ (W.GetHashCode() >> 1);
        }

        public override bool Equals(object other)
        {
            if (!(other is Quaternion))
                return false;
            return Equals((Quaternion)other);
        }

        public bool Equals(Quaternion other)
        {
            return X.Equals(other.X) && Y.Equals(other.Y) && Z.Equals(other.Z) && W.Equals(other.W);
        }

        public static float Dot(Quaternion a, Quaternion b)
        {
            return a.X * b.X + a.Y * b.Y + a.Z * b.Z + a.W * b.W;
        }

        private static bool IsEqualUsingDot(float dot)
        {
            return dot > 1.0f - kEpsilon;
        }

        public static bool operator ==(Quaternion lhs, Quaternion rhs)
        {
            return IsEqualUsingDot(Dot(lhs, rhs));
        }

        public static bool operator !=(Quaternion lhs, Quaternion rhs)
        {
            return !(lhs == rhs);
        }

        private const float kEpsilon = 0.000001F;
    }
}

"""



from math import sqrt


class Vector3:
    X: float
    Y: float
    Z: float

    def __init__(self, *args):
        if len(args) == 3 or len(args) == 1 and isinstance(args[0], (tuple, list)):
            self.X, self.Y, self.Z = args
        elif len(args) == 1:
            # dirty patch for Vector4
            self.__dict__ = args[0].__dict__

    def length(self) -> float:
        return sqrt(self.X ** 2 + self.Y ** 2 + self.Z ** 2)

    def normalize(self):
        length = self.length()

        invNorm = 1.0 / length
        self.X *= invNorm
        self.Y *= invNorm
        self.Z *= invNorm


"""
using System;
using System.Runtime.InteropServices;

namespace AssetStudio
{
	[StructLayout(LayoutKind.Sequential, Pack = 4)]
	public struct Vector3 : IEquatable<Vector3>
	{
		public float X;
		public float Y;
		public float Z;

		public Vector3(float x, float y, float z)
		{
			X = x;
			Y = y;
			Z = z;
		}

		public float this[int index]
		{
			get
			{
				switch (index)
				{
					case 0: return X;
					case 1: return Y;
					case 2: return Z;
					default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Vector3 index!");
				}
			}

			set
			{
				switch (index)
				{
					case 0: X = value; break;
					case 1: Y = value; break;
					case 2: Z = value; break;
					default: throw new ArgumentOutOfRangeException(nameof(index), "Invalid Vector3 index!");
				}
			}
		}

		public override int GetHashCode()
		{
			return X.GetHashCode() ^ (Y.GetHashCode() << 2) ^ (Z.GetHashCode() >> 2);
		}

		public override bool Equals(object other)
		{
			if (!(other is Vector3))
				return false;
			return Equals((Vector3)other);
		}

		public bool Equals(Vector3 other)
		{
			return X.Equals(other.X) && Y.Equals(other.Y) && Z.Equals(other.Z);
		}

		public void Normalize()
		{
			var length = Length();
			if (length > kEpsilon)
			{
				var invNorm = 1.0f / length;
				X *= invNorm;
				Y *= invNorm;
				Z *= invNorm;
			}
			else
			{
				X = 0;
				Y = 0;
				Z = 0;
			}
		}

		public float Length()
		{
			return (float)Math.Sqrt(LengthSquared());
		}

		public float LengthSquared()
		{
			return X * X + Y * Y + Z * Z;
		}

		public static Vector3 Zero => new Vector3();

		public static Vector3 operator +(Vector3 a, Vector3 b)
		{
			return new Vector3(a.X + b.X, a.Y + b.Y, a.Z + b.Z);
		}

		public static Vector3 operator -(Vector3 a, Vector3 b)
		{
			return new Vector3(a.X - b.X, a.Y - b.Y, a.Z - b.Z);
		}

		public static Vector3 operator -(Vector3 a)
		{
			return new Vector3(-a.X, -a.Y, -a.Z);
		}

		public static Vector3 operator *(Vector3 a, float d)
		{
			return new Vector3(a.X * d, a.Y * d, a.Z * d);
		}

		public static Vector3 operator *(float d, Vector3 a)
		{
			return new Vector3(a.X * d, a.Y * d, a.Z * d);
		}

		public static Vector3 operator /(Vector3 a, float d)
		{
			return new Vector3(a.X / d, a.Y / d, a.Z / d);
		}

		public static bool operator ==(Vector3 lhs, Vector3 rhs)
		{
			return (lhs - rhs).LengthSquared() < kEpsilon * kEpsilon;
		}

		public static bool operator !=(Vector3 lhs, Vector3 rhs)
		{
			return !(lhs == rhs);
		}

		public static implicit operator Vector2(Vector3 v)
		{
			return new Vector2(v.X, v.Y);
		}

		public static implicit operator Vector4(Vector3 v)
		{
			return new Vector4(v.X, v.Y, v.Z, 0.0F);
		}

		private const float kEpsilon = 0.00001F;
	}
}

"""



from .Object import Object


class BuildSettings(Object):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.levels = reader.read_string_array()
        self.has_render_texture = reader.read_boolean()
        self.has_pro_version = reader.read_boolean()
        self.has_publishing_rights = reader.read_boolean()
        self.has_shadows = reader.read_boolean()
        self.version = reader.read_aligned_string()



import io
import struct

from .math import Color, Matrix4x4, Quaternion, Vector2, Vector3, Vector4, Rectangle


class EndianBinaryWriter:
    endian: str
    Length: int
    Position: int
    stream: io.BufferedReader

    def __init__(self, input_=b"", endian=">"):
        if isinstance(input_, (bytes, bytearray)):
            self.stream = io.BytesIO(input_)
            self.stream = input_
        else:
            raise ValueError("Invalid input type - %s." % type(input_))
        self.endian = endian
        self.Position = self.stream.tell()

    @property
    def bytes(self):
        self.stream.seek(0)
        return self.stream.read()

    @property
    def Length(self) -> int:
        pos = self.stream.tell()
        self.stream.seek(0, 2)
        l = self.stream.tell()
        self.stream.seek(pos)
        return l

    def dispose(self):
        self.stream.close()
        pass

    def write(self, *args):
        if self.Position != self.stream.tell():
            self.stream.seek(self.Position)
        self.stream.write(*args)
        self.Position = self.stream.tell()

    def write_byte(self, value: int):
        self.write(struct.pack(self.endian + "b", value))

    def write_u_byte(self, value: int):
        self.write(struct.pack(self.endian + "B", value))

    def write_bytes(self, value: bytes):
        return self.write(value)

    def write_short(self, value: int):
        self.write(struct.pack(self.endian + "h", value))

    def write_int(self, value: int):
        self.write(struct.pack(self.endian + "i", value))

    def write_long(self, value: int):
        self.write(struct.pack(self.endian + "q", value))

    def write_u_short(self, value: int):
        self.write(struct.pack(self.endian + "H", value))

    def write_u_int(self, value: int):
        self.write(struct.pack(self.endian + "I", value))

    def write_u_long(self, value: int):
        self.write(struct.pack(self.endian + "Q", value))

    def write_float(self, value: float):
        self.write(struct.pack(self.endian + "f", value))

    def write_double(self, value: float):
        self.write(struct.pack(self.endian + "d", value))

    def write_boolean(self, value: bool):
        self.write(struct.pack(self.endian + "?", value))

    """
	def write_string(self, size=None, encoding="utf-8") -> str:
		if size is None:
			ret = self.write_string_to_null()
		else:
			ret = struct.unpack(f"{self.endian}{size}is", self.read(size))[0]
		try:
			return ret.decode(encoding)
		except UnicodeDecodeError:
			return ret
	"""

    def write_string_to_null(self, value: str):
        self.write(value.encode("utf8"))
        self.write(b"\0")

    def write_aligned_string(self, value: str):
        self.write_int(len(value))
        self.write(value.encode("utf8", "backslashreplace"))
        self.align_stream(4)

    def align_stream(self, alignment=4):
        pos = self.stream.tell()
        mod = pos % alignment
        if mod != 0:
            self.write(b"\0" * (alignment - mod))

    def write_quaternion(self, value: Quaternion):
        self.write_float(value.X)
        self.write_float(value.Y)
        self.write_float(value.Z)
        self.write_float(value.W)

    def write_vector2(self, value: Vector2):
        self.write_float(value.X)
        self.write_float(value.Y)

    def write_vector3(self, value: Vector3):
        self.write_float(value.X)
        self.write_float(value.Y)
        self.write_float(value.Z)

    def write_vector4(self, value: Vector4):
        self.write_float(value.X)
        self.write_float(value.Y)
        self.write_float(value.Z)
        self.write_float(value.W)

    def write_rectangle_f(self, value: Rectangle):
        self.write_float(value.x)
        self.write_float(value.y)
        self.write_float(value.width)
        self.write_float(value.height)

    def write_color4(self, value: Color):
        self.write_float(value.R)
        self.write_float(value.G)
        self.write_float(value.B)
        self.write_float(value.A)

    def write_matrix(self, value: Matrix4x4):
        for val in value.M:
            self.write_float(val)

    def write_array(self, command, value: list, write_length: bool = True):
        if write_length:
            self.write_int(len(value))
        for val in value:
            command(val)

    def write_boolean_array(self, value: list):
        self.write_array(self.write_boolean, value)

    def write_u_short_array(self, value: list):
        self.write_array(self.write_u_short, value)

    def write_int_array(self, value: list, write_length: bool = False):
        return self.write_array(self.write_int, value, write_length)

    def write_u_int_array(self, value: list, write_length: bool = False):
        return self.write_array(self.write_u_int, value, write_length)

    def write_float_array(self, value: list, write_length: bool = False):
        return self.write_array(self.write_float, value, write_length)

    def write_string_array(self, value: list):
        self.write_array(self.write_aligned_string, value)

    def write_vector2_array(self, value: list):
        self.write_array(self.write_vector2, value)

    def write_vector4_array(self, value: list):
        self.write_array(self.write_vector4, value)

    def write_matrix_array(self, value: list):
        self.write_array(self.write_matrix, value)



from ..enums import BuildTarget
from ..helpers import TypeTreeHelper
from ..streams import EndianBinaryWriter, EndianBinaryReader


class Object:
    type_tree: dict

    def __init__(self, reader):
        self.reader = reader
        self.assets_file = reader.assets_file
        self.type = reader.type
        self.path_id = reader.path_id
        self.version = reader.version
        self.build_type = reader.build_type
        self.platform = reader.platform
        self.serialized_type = reader.serialized_type
        self.byte_size = reader.byte_size
        self.environment = reader.assets_file.environment

        if self.platform == BuildTarget.NoTarget:
            self._object_hide_flags = reader.read_u_int()

        self.container = (
            self.assets_file._container[self.path_id]
            if self.path_id in self.assets_file._container
            else None
        )

        self.reader.reset()
        if type(self) == Object:
            self.__dict__.update(self.read_type_tree())

    def has_struct_member(self, name: str) -> bool:
        return self.serialized_type.m_Nodes and any(
            [x.name == name for x in self.serialized_type.m_Nodes]
        )

    def dump(self) -> str:
        self.reader.reset()
        if getattr(self.serialized_type, "nodes", None):
            sb = []
            TypeTreeHelper(self.reader).read_type_string(sb, self.serialized_type.nodes)
            return "".join(sb)
        return ""

    def read_type_tree(self) -> dict:
        self.reader.reset()
        if self.serialized_type.nodes:
            self.type_tree = TypeTreeHelper(self.reader).read_value(
                self.serialized_type.nodes, 0
            )
        else:
            self.type_tree = {}
        return self.type_tree

    def get_raw_data(self) -> bytes:
        self.reader.reset()
        return self.reader.read_bytes(self.byte_size)

    def save(self, writer: EndianBinaryWriter, intern_call=False):
        if intern_call:
            if self.platform == BuildTarget.NoTarget:
                writer.write_u_int(self._object_hide_flags)

    def _save(self, writer):
        # the reader is actually an ObjectReader,
        # the data value is written back into the asset
        self.reader.data = writer.bytes

    def __getattr__(self, item):
        """
        If item not found in __dict__, read type_tree and check if it is in there.
        """
        self.read_type_tree()
        if item in self.type_tree:
            return self.type_tree[item]

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)



from .File import File
from .SerializedFile import SerializedFile
from ..enums import FileType
from ..helpers import CompressionHelper, ImportHelper
from ..streams import EndianBinaryReader, EndianBinaryWriter


class BundleFile(File):
    format: int
    version_player: str
    version_engine: str

    def __init__(self, reader: EndianBinaryReader, environment=None):
        self.files = {}
        self.signature = reader.read_string_to_null()
        self.format = reader.read_int()
        self.version_player = reader.read_string_to_null()
        self.version_engine = reader.read_string_to_null()

        if self.format == 6:
            self.read_format_6(reader, self.signature != "UnityFS")

        elif self.signature in [
            "UnityWeb",
            "UnityRaw",
            "\xFA\xFA\xFA\xFA\xFA\xFA\xFA\xFA",
        ]:
            if self.format < 6:
                bundle_size = reader.read_int()

            dummy2 = reader.read_short()
            offset = reader.read_short()

            if self.signature in ["UnityWeb", "\xFA\xFA\xFA\xFA\xFA\xFA\xFA\xFA"]:
                dummy3 = reader.read_int()
                lzma_chunks = reader.read_int()
                reader.Position = reader.Position + (lzma_chunks - 1) * 8
                lzma_size = reader.read_int()
                stream_size = reader.read_int()

                reader.Position = offset
                lzma_buffer = reader.read_bytes(lzma_size)
                data_reader = EndianBinaryReader(
                    CompressionHelper.decompress_lzma(lzma_buffer)
                )
                self.get_assets_files(data_reader, 0)

            elif self.signature == "UnityRaw":
                reader.Position = offset
                self.get_assets_files(reader, offset)

        else:
            raise ValueError(
                f"unknown combination of format ({self.format}) & signature ({self.signature})"
            )

        for name, item in self.files.items():
            # check for serialized files
            if name.endswith((".resS", ".resource", ".config", ".xml", ".dat")):
                environment.resources[name] = item
            else:
                typ, _ = ImportHelper.check_file_type(item)
                if typ == FileType.AssetsFile:
                    item.Position = 0
                    sf = SerializedFile(item, environment, self)
                    sf.flag = item.flag
                    self.files[name] = sf
                    environment.assets[name] = sf
                else:
                    environment.resources[name] = item

    def get_assets_files(self, reader: EndianBinaryReader, offset):
        file_count = reader.read_int()
        files = [
            # name, offset, size
            (reader.read_string_to_null(), reader.read_int(), reader.read_int())
            for _ in range(file_count)
        ]
        for name, f_offset, size in files:
            reader.Position = offset + f_offset
            self.files[name] = EndianBinaryReader(reader.read(size))

    def read_format_6(self, reader: EndianBinaryReader, padding):
        bundle_size = reader.read_long()
        compressed_size = reader.read_int()
        uncompressed_size = reader.read_int()
        flag = reader.read_int()

        if padding:
            reader.read_byte()

        if (flag & 0x80) != 0:  # at end of file
            position = reader.Position
            reader.Position = reader.Length - compressed_size
            block_info_bytes = reader.read_bytes(compressed_size)
            reader.Position = position
        else:
            block_info_bytes = reader.read_bytes(compressed_size)

        switch = flag & 0x3F
        if switch == 1:  # LZMA
            blocks_info_data = CompressionHelper.decompress_lzma(block_info_bytes)
        elif switch in [2, 3]:  # LZ4, LZ4HC
            blocks_info_data = CompressionHelper.decompress_lz4(
                block_info_bytes, uncompressed_size
            )
        # elif switch == 4: #LZHAM:
        else:  # no compression
            blocks_info_data = block_info_bytes

        blocks_info_reader = EndianBinaryReader(blocks_info_data)
        blocks_info_reader.Position = 0x10
        block_count = blocks_info_reader.read_int()
        block_infos = [BlockInfo(blocks_info_reader) for _ in range(block_count)]

        data = []
        for block_info in block_infos:
            switch = block_info.flag & 0x3F

            if switch == 1:  # LZMA
                data.append(
                    CompressionHelper.decompress_lzma(
                        reader.read(block_info.compressed_size)
                    )
                )
            elif switch in [2, 3]:  # LZ4, LZ4HC
                data.append(
                    CompressionHelper.decompress_lz4(
                        reader.read(block_info.compressed_size),
                        block_info.uncompressed_size,
                    )
                )
            # elif switch == 4: #LZHAM:
            else:  # no compression
                data.append(reader.read(block_info.compressed_size))

        data_stream = EndianBinaryReader(b"".join(data))
        entry_info_count = blocks_info_reader.read_int()

        for _ in range(entry_info_count):
            offset = blocks_info_reader.read_long()
            size = blocks_info_reader.read_long()
            flag = blocks_info_reader.read_int()
            name = blocks_info_reader.read_string_to_null()
            data_stream.Position = offset

            item = EndianBinaryReader(data_stream.read(size))
            item.flag = flag
            self.files[name] = item

    def save(self):
        # file_header
        #     signature	(string_to_null)
        #     format		(int)
        #     version_player	(string_to_null)
        #     version_engine	(string_to_null)
        writer = EndianBinaryWriter()

        writer.write_string_to_null(self.signature)
        writer.write_int(self.format)
        writer.write_string_to_null(self.version_player)
        writer.write_string_to_null(self.version_engine)

        if self.format == 6:  # WORKS
            self.save_format_6(writer, self.signature != "UnityFS")
        else:  # WIP
            raise NotImplementedError("Not Implemented")

        return writer.bytes

    def save_format_6(self, writer: EndianBinaryWriter, padding=False):
        # 192 - uncompressed
        # 194 - lz4
        block_info_flag = 64
        # 0/64 - uncompressed
        # 1 - lzma
        # 2 - lz4
        # 3 - lz4hc - not implemented
        # 4 - lzham - not implemented
        data_flag = 64
        # header:
        #     bundle_size		(long)
        #     compressed_size	(int)
        #     uncompressed_size	(int)
        #     flag			    (int)
        #     ?padding?		    (bool)
        #   This will be written at the end,
        #   because the size can only be calculated after the data compression,

        # block_info:
        #     *flag & 0x80 ? at the end : right after header
        #     *decompression via flag & 0x3F
        #     *read compressed_size -> uncompressed_size
        #     0x10 offset
        #     *read blocks infos of the data stream
        #     count			(int)
        #     (
        #         uncompressed_size(uint)
        #         compressed_size (uint)
        #         flag(short)
        #     )
        #     *decompression via info.flag & 0x3F

        #     *afterwards the file positions
        #     file_count		(int)
        #     (
        #         offset	(long)
        #         size		(long)
        #         flag		(int)
        #         name		(string_to_null)
        #     )

        # file list & file data
        data_writer = EndianBinaryWriter()
        files = [
            (
                name,
                f.flag,
                data_writer.write_bytes(
                    f.bytes if isinstance(f, EndianBinaryReader) else f.save()
                ),
            )
            for name, f in self.files.items()
        ]

        file_data = data_writer.bytes
        data_writer.dispose()
        uncompressed_data_size = len(file_data)

        # compress the data
        switch = data_flag & 0x3F
        if switch == 1:  # LZMA
            file_data = CompressionHelper.compress_lzma(file_data)
        elif switch in [2, 3]:  # LZ4, LZ4HC
            file_data = CompressionHelper.compress_lz4(file_data)
        elif switch == 4:  # LZHAM
            raise NotImplementedError
        # else no compression - data stays the same
        compressed_data_size = len(file_data)

        # write the block_info
        block_writer = EndianBinaryWriter(b"\x00" * 0x10)
        # data block info
        # block count
        block_writer.write_int(1)
        # uncompressed size
        block_writer.write_u_int(uncompressed_data_size)
        # compressed size
        block_writer.write_u_int(compressed_data_size)
        # flag
        block_writer.write_short(data_flag)

        # file block info
        # file count
        block_writer.write_int(len(files))
        offset = 0
        for f_name, f_flag, f_len in files:
            # offset
            block_writer.write_long(offset)
            # size
            block_writer.write_long(f_len)
            offset += f_len
            # flag
            block_writer.write_int(f_flag)
            # name
            block_writer.write_string_to_null(f_name)

        # compress the block data
        block_data = block_writer.bytes
        block_writer.dispose()

        uncompressed_block_data_size = len(block_data)

        switch = block_info_flag & 0x3F
        if switch == 1:  # LZMA
            block_data = CompressionHelper.compress_lzma(block_data)
        elif switch in [2, 3]:  # LZ4, LZ4HC
            block_data = CompressionHelper.compress_lz4(block_data)
        elif switch == 4:  # LZHAM
            raise NotImplementedError

        compressed_block_data_size = len(block_data)

        # write the header info
        writer.write_long(
            writer.Length
            + 8
            + 4
            + 4
            + 4
            + (1 if padding else 0)
            + compressed_block_data_size
            + compressed_data_size
        )
        writer.write_int(compressed_block_data_size)
        writer.write_int(uncompressed_block_data_size)
        writer.write_int(block_info_flag)
        if padding:
            writer.write_boolean(padding)

        if (block_info_flag & 0x80) != 0:  # at end of file
            writer.write(file_data)
            writer.write(block_data)
        else:
            writer.write(block_data)
            writer.write(file_data)


class BlockInfo:
    compressed_size: int
    uncompressed_size: int
    flag: int

    def __init__(self, reader: EndianBinaryReader):
        self.uncompressed_size = reader.read_u_int()
        self.compressed_size = reader.read_u_int()
        self.flag = reader.read_short()



from enum import IntEnum

from .Mesh import BoneWeights4, SubMesh, VertexData
from .NamedObject import NamedObject
from .PPtr import PPtr
from ..export import SpriteHelper


class Sprite(NamedObject):
    @property
    def image(self):
        return SpriteHelper.get_image_from_sprite(self)

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        self.m_Rect = reader.read_rectangle_f()
        self.m_Offset = reader.read_vector2()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 5):  # 4.5 and up
            self.m_Border = reader.read_vector4()

        self.m_PixelsToUnits = reader.read_float()
        if (
            version[0] > 5
            or (version[0] == 5 and version[1] > 4)
            or (version[0] == 5 and version[1] == 4 and version[2] >= 2)
            or (
                version[0] == 5
                and version[1] == 4
                and version[2] == 1
                and self.build_type.IsPatch
                and version[3] >= 3
            )
        ):  # 5.4.1p3 and up
            self.m_Pivot = reader.read_vector2()

        self.m_Extrude = reader.read_u_int()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 3):  # 5.3 and up
            self.m_IsPolygon = reader.read_boolean()
            reader.align_stream()

        if version[0] >= 2017:  # 2017 and up
            first = reader.read_bytes(16)  # GUID
            second = reader.read_long()
            self.m_RenderDataKey = (first, second)
            self.m_AtlasTags = reader.read_string_array()
            self.m_SpriteAtlas = PPtr(reader)  # SpriteAtlas

        self.m_RD = SpriteRenderData(reader)

        if version[0] >= 2017:  # 2017 and up
            m_PhysicsShapeSize = reader.read_int()
            self.m_PhysicsShape = [
                reader.read_vector2_array() for _ in range(m_PhysicsShapeSize)
            ]


# vector m_Bones 2018 and up


class SecondarySpriteTexture:
    def __init__(self, reader):
        self.texture = PPtr(reader)  # Texture2D
        self.name = reader.read_string_to_null()


class SpritePackingRotation(IntEnum):
    kSPRNone = (0,)
    kSPRFlipHorizontal = (1,)
    kSPRFlipVertical = (2,)
    kSPRRotate180 = (3,)
    kSPRRotate90 = 4


class SpritePackingMode(IntEnum):
    kSPMTight = (0,)
    kSPMRectangle = 1


class SpriteSettings:
    def __init__(self, reader):
        settingsRaw = reader.read_u_int()
        self.packed = settingsRaw & 1  # 1
        self.packingMode = SpritePackingMode((settingsRaw >> 1) & 1)  # 1
        self.packingRotation = SpritePackingRotation((settingsRaw >> 2) & 0xF)  # 4


# meshType = (settingsRaw >> 6) & 1 #1
# reserved


class SpriteVertex:
    def __init__(self, reader):
        version = reader.version
        self.pos = reader.read_vector3()
        if version[0] < 4 or (version[0] == 4 and version[1] <= 3):  # 4.3 and down
            self.uv = reader.read_vector2()


class SpriteRenderData:
    def __init__(self, reader):
        version = reader.version

        self.texture = PPtr(reader)  # Texture2D
        if version[0] > 5 or (version[0] == 5 and version[1] >= 2):  # 5.2 and up
            self.alphaTexture = PPtr(reader)  # Texture2D

        if version[0] >= 2019:  # 2019 and up
            secondaryTexturesSize = reader.read_int()
            self.secondaryTextures = [
                SecondarySpriteTexture(reader) for _ in range(secondaryTexturesSize)
            ]

        if version[0] > 5 or (version[0] == 5 and version[1] >= 6):  # 5.6 and up
            m_SubMeshesSize = reader.read_int()
            self.m_SubMeshes = [SubMesh(reader) for _ in range(m_SubMeshesSize)]
            self.m_IndexBuffer = reader.read_bytes(reader.read_int())
            reader.align_stream()
            self.m_VertexData = VertexData(reader)
        else:
            verticesSize = reader.read_int()
            self.vertices = [SpriteVertex(reader) for _ in range(verticesSize)]
            self.indices = reader.read_u_short_array()
            reader.align_stream()

        if version[0] >= 2018:  # 2018 and up
            self.m_Bindpose = reader.read_matrix_array()
            if version[0] == 2018 and version[1] < 2:  # 2018.2 down
                m_SourceSkinSize = reader.read_int()
                self.m_SourceSkin = [BoneWeights4(reader)]

        self.textureRect = reader.read_rectangle_f()
        self.textureRectOffset = reader.read_vector2()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 6):  # 5.6 and up
            self.atlasRectOffset = reader.read_vector2()

        self.settingsRaw = SpriteSettings(reader)
        if version[0] > 4 or (version[0] == 4 and version[1] >= 5):  # 4.5 and up
            self.uvTransform = reader.read_vector4()

        if version[0] >= 2017:  # 2017 and up
            self.downscaleMultiplier = reader.read_float()



class File(object):
    name: str
    files: dict
    signature: str
    compression: str
    # parent: File

    def keys(self):
        return self.files.keys()

    def items(self):
        return self.files.items()

    def values(self):
        return self.files.values()

    def __getitem__(self, item):
        return self.files[item]

    def __repr__(self):
        return f"<{self.__class__.__name__}>"



from .Component import Component


class Behaviour(Component):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.enabled = reader.read_byte()
        self._dummy = reader.align_stream()



import io
import struct

from ..math import Color, Matrix4x4, Quaternion, Vector2, Vector3, Vector4, Rectangle


class EndianBinaryReader:
    endian: str
    Length: int
    Position: int
    stream: io.BufferedReader

    def __init__(self, input_, endian=">"):
        if isinstance(input_, (bytes, bytearray)):
            self.stream = io.BytesIO(input_)
        elif isinstance(input_, (io.BytesIO, io.BufferedReader)):
            self.stream = input_
        else:
            # test if input is a streamable object
            try:
                p = input_.tell()
                input_.read(1)
                input_.seek(p)
                assert p == input_.tell()
                self.stream = input_
            except:
                raise ValueError("Invalid input type - %s." % type(input_))

        self.endian = endian
        self.Length = self.stream.seek(0, 2)
        self.Position = 0

    def get_position(self):
        return self.stream.tell()

    def set_position(self, value):
        self.stream.seek(value)

    Position = property(get_position, set_position)

    @property
    def bytes(self):
        last_pos = self.Position
        self.Position = 0
        ret = self.read()
        self.Position = last_pos
        return ret

    def __add__(self, value):
        if isinstance(value, (bytes, bytearray)):
            old_pos = self.Position
            self.Position = self.Length
            self.stream.write(value)
            self.Length += len(value)
            self.Position = old_pos
        else:
            raise ValueError("Invalid Input")

    def dispose(self):
        self.stream.close()
        pass

    def read(self, *args):
        return self.stream.read(*args)

    def read_byte(self) -> int:
        return struct.unpack(self.endian + "b", self.read(1))[0]

    def read_u_byte(self) -> int:
        return struct.unpack(self.endian + "B", self.read(1))[0]

    def read_bytes(self, num) -> bytes:
        return self.read(num)

    def read_short(self) -> int:
        return struct.unpack(self.endian + "h", self.read(2))[0]

    def read_int(self) -> int:
        return struct.unpack(self.endian + "i", self.read(4))[0]

    def read_long(self) -> int:
        return struct.unpack(self.endian + "q", self.read(8))[0]

    def read_u_short(self) -> int:
        return struct.unpack(self.endian + "H", self.read(2))[0]

    def read_u_int(self) -> int:
        return struct.unpack(self.endian + "I", self.read(4))[0]

    def read_u_long(self) -> int:
        return struct.unpack(self.endian + "Q", self.read(8))[0]

    def read_float(self) -> float:
        return struct.unpack(self.endian + "f", self.read(4))[0]

    def read_double(self) -> float:
        return struct.unpack(self.endian + "d", self.read(8))[0]

    def read_boolean(self) -> bool:
        return bool(struct.unpack(self.endian + "?", self.read(1))[0])

    def read_string(self, size=None, encoding="utf-8") -> str:
        if size is None:
            ret = self.read_string_to_null()
        else:
            ret = struct.unpack(f"{self.endian}{size}is", self.read(size))[0]
        try:
            return ret.decode(encoding)
        except UnicodeDecodeError:
            return ret

    def read_string_to_null(self, max_length=32767) -> str:
        ret = []
        c = b""
        while c != b"\0" and len(ret) < max_length and self.Position != self.Length:
            ret.append(c)
            c = self.read(1)
            if not c:
                raise ValueError("Unterminated string: %r" % ret)
        return b"".join(ret).decode("utf8", "replace")

    def read_aligned_string(self):
        length = self.read_int()
        if 0 < length <= self.Length - self.Position:
            string_data = self.read_bytes(length)
            result = string_data.decode("utf8", "backslashreplace")
            self.align_stream(4)
            return result
        return ""

    def align_stream(self, alignment=4):
        pos = self.Position
        mod = pos % alignment
        if mod != 0:
            self.Position += alignment - mod

    def read_quaternion(self):
        return Quaternion(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_vector2(self):
        return Vector2(self.read_float(), self.read_float())

    def read_vector3(self):
        return Vector3(self.read_float(), self.read_float(), self.read_float())

    def read_vector4(self):
        return Vector4(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_rectangle_f(self):
        return Rectangle(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_color4(self):
        return Color(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_matrix(self):
        return Matrix4x4(self.read_float_array(16))

    def read_array(self, command, length: int):
        return [command() for i in range(length)]

    def read_boolean_array(self):
        return self.read_array(self.read_boolean, self.read_int())

    def read_u_short_array(self):
        return self.read_array(self.read_u_short, self.read_int())

    def read_int_array(self, length=0):
        return self.read_array(self.read_int, length if length else self.read_int())

    def read_u_int_array(self, length=0):
        return self.read_array(self.read_u_int, length if length else self.read_int())

    def read_float_array(self, length=0):
        return self.read_array(self.read_float, length if length else self.read_int())

    def read_string_array(self):
        return self.read_array(self.read_aligned_string, self.read_int())

    def read_vector2_array(self):
        return self.read_array(self.read_vector2, self.read_int())

    def read_vector4_array(self):
        return self.read_array(self.read_vector4, self.read_int())

    def read_matrix_array(self):
        return self.read_array(self.read_matrix, self.read_int())

    def read_u_int_array_array(self):
        return self.read_array(self.read_u_int_array, self.read_int())



from .NamedObject import NamedObject
from .PPtr import PPtr


class Material(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        self.m_Shader = PPtr(reader)  # Shader
        if version[0] == 4 and version[1] >= 1:  # 4.x
            self.m_ShaderKeywords = reader.read_string_array()

        if version[0] >= 5:  # 5.0 and up
            self.m_ShaderKeywords = reader.read_aligned_string()
            self.m_LightmapFlags = reader.read_u_int()

        if version[0] > 5 or (version[0] == 5 and version[1] >= 6):  # 5.6 and up
            self.m_EnableInstancingVariants = reader.read_boolean()
            # var m_DoubleSidedGI = a_Stream.read_boolean() //2017 and up
            reader.align_stream()

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_CustomRenderQueue = reader.read_int()

        if version[0] > 5 or (version[0] == 5 and version[1] >= 1):  # 5.1 and up
            stringTagMapSize = reader.read_int()
            self.stringTagMap = {}
            for _ in range(stringTagMapSize):
                first = reader.read_aligned_string()
                second = reader.read_aligned_string()
                self.stringTagMap[first] = second

        if version[0] > 5 or (version[0] == 5 and version[1] >= 6):  # 5.6 and up
            self.disabledShaderPasses = reader.read_string_array()

        self.m_SavedProperties = UnityPropertySheet(reader)


class UnityTexEnv:
    def __init__(self, reader):
        self.m_Texture = PPtr(reader)  # Texture
        self.m_Scale = reader.read_vector2()
        self.m_Offset = reader.read_vector2()


class UnityPropertySheet:
    def __init__(self, reader):
        m_TexEnvsSize = reader.read_int()
        self.m_TexEnvs = {}
        for i in range(m_TexEnvsSize):
            key = reader.read_aligned_string()
            self.m_TexEnvs[key] = UnityTexEnv(reader)

        m_FloatsSize = reader.read_int()
        self.m_Floats = {}
        for i in range(m_FloatsSize):
            key = reader.read_aligned_string()
            self.m_Floats[key] = reader.read_float()

        m_ColorsSize = reader.read_int()
        self.m_Colors = {}
        for i in range(m_ColorsSize):
            key = reader.read_aligned_string()
            self.m_Colors[key] = reader.read_color4()



from .Behaviour import Behaviour
from .PPtr import PPtr


class MonoBehaviour(Behaviour):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.script = PPtr(reader)
        self.name = reader.read_aligned_string()

        self.read_type_tree()



from .EditorExtension import EditorExtension
from .PPtr import PPtr


class Component(EditorExtension):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.GameObject = PPtr(reader)  # GameObject



from .Vector4 import Vector4


class Color:
    R: float
    G: float
    B: float
    A: float

    def __init__(self, r: float, g: float, b: float, a: float):
        self.R = r
        self.G = g
        self.B = b
        self.A = a

    def GetHashCode(self):
        return Vector4(self).GetHashCode()

    def __eq__(self, other):
        if isinstance(other, Color):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __add__(self, other):
        return Color(
            self.R + other.R, self.G + other.G, self.B + other.B, self.A + other.A
        )

    def __sub__(self, other):
        return Color(
            self.R - other.R, self.G - other.G, self.B - other.B, self.A - other.A
        )

    def __mul__(self, other):
        if isinstance(other, Color):
            return Color(
                self.R * other.R, self.G * other.G, self.B * other.B, self.A * other.A
            )
        else:
            return Color(self.R * other, self.G * other, self.B * other, self.A * other)

    def __div__(self, other):
        if isinstance(other, Color):
            return Color(
                self.R / other.R, self.G / other.G, self.B / other.B, self.A / other.A
            )
        else:
            return Color(self.R / other, self.G / other, self.B / other, self.A / other)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def Vector4(self):
        return Vector4(self.R, self.G, self.B, self.A)



from .NamedObject import NamedObject
from ..enums import AudioType, AudioCompressionFormat, AUDIO_TYPE_EXTEMSION
from ..export import AudioClipConverter
from ..helpers.ResourceReader import get_resource_data


class AudioClip(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_Source = ""
        version = self.version
        if version[0] < 5:  #
            self.m_Format = reader.read_int()
            self.m_Type = AudioType(reader.read_int())
            self.m_3D = reader.read_boolean()
            self.m_UseHardware = reader.read_boolean()
            reader.align_stream()

            if version[0] >= 4 or (version[0] == 3 and version[1] >= 2):  # 3.2.0 to 5
                self.m_Stream = reader.read_int()
                m_Size = reader.read_int()
                tsize = m_Size + 4 - m_Size % 4 if (m_Size % 4 != 0) else m_Size
                if reader.byte_size + reader.byte_start - reader.Position != tsize:
                    m_Offset = reader.read_int()
                    self.m_Source = self.assets_file.full_name + ".resS"
            else:
                m_Size = reader.read_int()

            self.extension = AUDIO_TYPE_EXTEMSION.get(self.m_Type, ".audioclip")
        else:
            self.m_LoadType = reader.read_int()
            self.m_Channels = reader.read_int()
            self.m_Frequency = reader.read_int()
            self.m_BitsPerSample = reader.read_int()
            self.m_Length = reader.read_float()
            self.m_IsTrackerFormat = reader.read_boolean()
            reader.align_stream()
            self.m_SubsoundIndex = reader.read_int()
            self.m_PreloadAudioData = reader.read_boolean()
            self.m_LoadInBackground = reader.read_boolean()
            self.m_Legacy3D = reader.read_boolean()
            reader.align_stream()
            self.m_Source = reader.read_aligned_string()
            m_Offset = reader.read_long()
            m_Size = reader.read_long()
            self.m_CompressionFormat = AudioCompressionFormat(reader.read_int())

            self.extension = AUDIO_TYPE_EXTEMSION.get(
                self.m_CompressionFormat, ".audioclip"
            )

        if self.m_Source:
            self.m_AudioData = get_resource_data(
                self.m_Source, self.assets_file, m_Offset, m_Size
            )
        else:
            self.m_AudioData = reader.read(m_Size)

    @property
    def samples(self) -> dict:
        return AudioClipConverter.extract_audioclip_samples(self)



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="UnityPy",
	packages=setuptools.find_packages(),
	version="1.4.1",
	author="K0lb3",
	description="A pythonic port of AssetStudio by Perfare",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/K0lb3/UnityPy",
	download_url="https://github.com/K0lb3/UnityPy/tarball/master",
	keywords=['Unity', 'unitypack', 'UnityPy', "unpack"],
	classifiers=[
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Topic :: Multimedia :: Graphics",
	],
	install_requires=[
		"lz4",
		"brotli",
		"Pillow",
		"fsb5",
		"texture2ddecoder"
	]
)



from ..streams import EndianBinaryReader


class RefInt:
    v: int

    def __init__(self, value):
        self._value = value

    def __add__(self, other):
        return self._value + other

    def __sub__(self, other):
        self._value -= other
        return self._value

    def __int__(self):
        return self._value

    def __getattr__(self, item):
        return self._value

    def __getitem__(self, item):
        return self._value

    def __setattr__(self, key, value):
        self.__dict__["_value"] = value

    def __setitem__(self, key, value):
        self._value = value

    def __mod__(self, other):
        return self._value % other

    def __ge__(self, other):
        return self._value >= other

    def __gt__(self, other):
        return self._value > other

    def __le__(self, other):
        return self._value <= other

    def __lt__(self, other):
        return self._value < other

    def __eq__(self, other):
        return self._value == other


def get_members(members: list, level: int, index: int) -> list:
    member2 = [members[0]]
    for i in range(index + 1, len(members)):
        member = members[i]
        if member.level <= level:
            return member2
        member2.append(member)
    return member2


class TypeTreeHelper:
    def __init__(self, reader: EndianBinaryReader):
        self.reader = reader
        self.READ = {
            "SInt8": self.reader.read_byte,
            "UInt8": self.reader.read_u_byte,
            "short": self.reader.read_short,
            "SInt16": self.reader.read_short,
            "unsigned short": self.reader.read_u_short,
            "UInt16": self.reader.read_u_short,
            "int": self.reader.read_int,
            "SInt32": self.reader.read_int,
            "unsigned int": self.reader.read_u_int,
            "UInt32": self.reader.read_u_int,
            "Type*": self.reader.read_u_int,
            "long long": self.reader.read_long,
            "SInt64": self.reader.read_long,
            "unsigned long long": self.reader.read_u_long,
            "UInt64": self.reader.read_u_long,
            "float": self.reader.read_float,
            "double": self.reader.read_double,
            "bool": self.reader.read_boolean,
        }

        self.READ2 = {
            "string": self.read_string,
            "map": self.read_map,
            "TypelessData": self.read_typeless_data,
        }

    def read_u_type(self, members: list) -> dict:
        i = RefInt(0)
        obj = {}
        while i.v < len(members):
            member = members[i.v]
            obj[member.name] = self.read_value(members, i)
            i.v += 1
        return obj

    def read_string(self, i, align, *args):
        value = self.reader.read_aligned_string()
        i.v += 3
        return value, align

    def read_map(self, i, align, members, level):
        if (members[i + 1].meta_flag & 0x4000) != 0:
            align = True
        size = self.reader.read_int()
        map_ = get_members(members, level, i)[4:]
        i.v += len(map_) + 3
        first = get_members(map_, map_[0].level, 0)
        second = map_[len(first) :]
        value = {}
        for j in range(size):
            tmp1 = RefInt(0)
            tmp2 = RefInt(0)
            v1 = self.read_value(
                first, tmp1
            )  # python reads the value first and then the key, so it has to be this way
            if isinstance(v1, dict):
                value[first] = tmp1
                value["values"] = self.read_value(second, tmp2)
            else:
                value[v1] = self.read_value(second, tmp2)
        return value, align

    def read_typeless_data(self, i, align, *args):
        size = self.reader.read_int()
        value = self.reader.read_bytes(size)
        i.v += 2
        return value, align

    def read_value(self, members: list, i) -> object:
        if type(i) != RefInt:
            i = RefInt(i)
        member = members[i.v]
        level = member.level
        var_type_str = member.type
        align = (member.meta_flag & 0x4000) != 0

        value = self.READ.get(var_type_str)
        if value:
            value = value()
        else:
            value = self.READ2.get(var_type_str)
            if value:
                value, align = value(i, align, members, level)
            else:
                if i != len(members) and members[i.v + 1].type == "Array":
                    if (members[i.v + 1].meta_flag & 0x4000) != 0:
                        align = True
                    size = self.reader.read_int()
                    vector = get_members(members, level, i)[3:]
                    i.v += len(vector) + 2
                    value = [self.read_value(vector, 0) for j in range(size)]
                else:
                    eclass = get_members(members, level, i)
                    eclass.pop(0)
                    i.v += len(eclass)
                    j = RefInt(0)
                    value = {}
                    while j < len(eclass):
                        classmember = eclass[j.v]
                        name = classmember.name
                        value[name] = self.read_value(eclass, j)
                        j.v += 1
        if align:
            self.reader.align_stream()
        return value

    def read_type_string(self, sb: list, members: list):
        i = RefInt(0)
        while i < len(members):
            self.read_string_value(sb, members, i)
            i.v += 1
        return sb

    def read_string_value(self, sb: list, members, i: RefInt):
        if type(i) != RefInt:
            i = RefInt(i)
        member = members[i.v]
        level = member.level
        var_type_str = member.type
        var_name_str = member.name
        append = True
        align = (member.meta_flag & 0x4000) != 0
        value = self.READ.get(var_type_str)
        if value:
            value = value()

        elif var_type_str == "string":
            append = False
            string = self.reader.read_aligned_string()
            sb.append(
                '{0}{1} {2} = "{3}"\r\n'.format(
                    "\t" * level, var_type_str, var_name_str, string
                )
            )
            i.v += 3

        elif var_type_str == "vector":
            if (members[i + 1].meta_flag & 0x4000) != 0:
                align = True
            append = False
            sb.append("{0}{1} {2}\r\n".format("\t" * level, var_type_str, var_name_str))
            sb.append("{0}{1} {2}\r\n".format("\t" * (level + 1), "Array", "Array"))
            size = self.reader.read_int()
            sb.append(
                "{0}{1} {2} = {3}\r\n".format("\t" * (level + 1), "int", "size", size)
            )
            vector = get_members(members, level, i.v)[3:]
            i.v += len(vector) + 2
            for j in range(size):
                sb.append("{0}[{1}]\r\n".format("\t" * (level + 2), j))
                tmp = RefInt(0)
                self.read_string_value(sb, vector, tmp)

        elif var_type_str == "map":
            if (members[i + 1].meta_flag & 0x4000) != 0:
                align = True
            append = False
            sb.append("{0}{1} {2}\r\n".format("\t" * level, var_type_str, var_name_str))
            sb.append("{0}{1} {2}\r\n".format("\t" * (level + 1), "Array", "Array"))
            size = self.reader.read_int()
            sb.append(
                "{0}{1} {2} = {3}\r\n".format("\t" * (level + 1), "int", "size", size)
            )
            map_ = get_members(members, level, i.v)[4:]
            i.v += len(map_) + 3
            first = get_members(map_, map_[0].level, 0)
            second = map_[len(first) :]
            for j in range(size):
                sb.append("{0}[{1}]\r\n".format("\t" * (level + 2), j))
                sb.append("{0}{1} {2}\r\n".format("\t" * (level + 2), "pair", "data"))
                tmp1 = RefInt(0)
                tmp2 = RefInt(0)
                self.read_string_value(sb, first, tmp1)
                self.read_string_value(sb, second, tmp2)

        elif var_type_str == "TypelessData":
            append = False
            size = self.reader.read_int()
            value = self.reader.read_bytes(size)
            i.v += 2
            sb.append("{0}{1} {2}\r\n".format("\t" * level, var_type_str, var_name_str))
            sb.append("{0}{1} {2} = {3}\r\n".format("\t" * level, "int", "size", size))

        else:
            if i != len(members) and members[i + 1].type == "Array":
                if (members[i + 1].meta_flag & 0x4000) != 0:
                    align = True
                append = False
                sb.append(
                    "{0}{1} {2}\r\n".format("\t" * level, var_type_str, var_name_str)
                )
                sb.append("{0}{1} {2}\r\n".format("\t" * (level + 1), "Array", "Array"))
                size = self.reader.read_int()
                sb.append(
                    "{0}{1} {2} = {3}\r\n".format(
                        "\t" * (level + 1), "int", "size", size
                    )
                )
                vector = get_members(members, level, i.v)
                i.v += len(vector) - 1
                vector = vector[3:]  # vector.RemoveRange(0, 3)
                for j in range(size):
                    sb.append("{0}[{1}]\r\n".format("\t" * (level + 2), j))
                    tmp = RefInt(0)
                    self.read_string_value(sb, vector, tmp)
            else:
                append = False
                sb.append(
                    "{0}{1} {2}\r\n".format("\t" * level, var_type_str, var_name_str)
                )
                eclass = get_members(members, level, i.v)
                eclass.pop(0)  # .RemoveAt(0)
                i.v += len(eclass)
                j = RefInt(0)
                while j < len(eclass):
                    self.read_string_value(sb, eclass, j)
                    j.v += 1

        if append:
            sb.append(
                "{0}{1} {2} = {3}\r\n".format(
                    "\t" * level, var_type_str, var_name_str, value
                )
            )
        if align:
            self.reader.align_stream()

        return sb


"""
	def write_u_type(self, obj: dict, members: list) -> bytes:
		stream = io.BytesIO()
		write = EndianBinaryReader(stream)
		for i, member in enumerate(members):
			var_name_str = member.name
			write_value(obj[var_name_str], members, write, RefInt(i))
		pos = stream.tell()
		stream.seek(0)
		ret = stream.read()
		stream.seek(pos)
		return ret
	
	def write_value(value: dict, members: list, write: EndianBinaryReader, i):
		member = members[i.v]
		level = member.level
		var_type_str = member.type
		align = (member.meta_flag & 0x4000) != 0
		if var_type_str == "SInt8":
			write.write_s_byte(value)
		elif var_type_str == "UInt8":
			write.write_byte(value)
		elif var_type_str in ["short", "SInt16"]:
			write.write_short(value)
		elif var_type_str in ["UInt16", "unsigned short"]:
			write.write_u_short(value)
		elif var_type_str in ["int", "SInt32"]:
			write.write_int(value)
		elif var_type_str in ["UInt32", "unsigned int", "Type*"]:
			write.write_u_int(value)
		elif var_type_str in ["long long", "SInt64"]:
			write.write_long(value)
		elif var_type_str in ["UInt64", "unsigned long long"]:
			write.write_u_long(value)
		elif var_type_str == "float":
			write.write_float(value)
		elif var_type_str == "double":
			write.write_double(value)
		elif var_type_str == "bool":
			write.write_bool(value)
		elif var_type_str == "string":
			write.write_aligned_string(value)
			i.v += 3
		elif var_type_str == "map":
			if (members[i + 1].meta_flag & 0x4000) != 0:
				align = True
			dic = list(value.items())
			size = len(dic)
			write.Write(size)
			map_ = get_members(members, level, i)
			i.v += len(map_) - 1
			map_ = map_[4:]  # map.RemoveRange(0, 4)
			first = get_members(map_, map_[0].level, 0)
			map_ = map_[len(first):]  # .RemoveRange(0, len(first))
			second = map_
			for j in range(size):
				tmp1 = RefInt(0)
				tmp2 = RefInt(0)
				write_value(dic[j][0], first, write, tmp1)  # key
				write_value(dic[j][1], second, write, tmp2)  # value
		elif var_type_str == "TypelessData":
			byts = bytes(value)
			size = len(byts)
			write.Write(size)
			write.Write(bytes)
			i.v += 2
		else:
			if i != len(members) and members[i + 1].type == "Array":  # Array
				if (members[i.v + 1].meta_flag & 0x4000) != 0:
					align = True
				list_ = value
				size = len(list_)
				write.Write(size)
				vector = get_members(members, level, i)
				i.v += len(vector) - 1
				vector = vector[3:]  # vector.RemoveRange(0, 3)
				for j in range(size):
					tmp = RefInt(0)
					write_value(list_[j], vector, write, tmp)
			else:  # Class
				eclass = get_members(members, level, i)
				eclass.pop(0)  # .RemoveAt(0)
				i.v += len(eclass)
				obj = value
				j = RefInt(0)
				while j < len(eclass):
					classmember = eclass[j.v]
					name = classmember.name
					write_value(obj[name], eclass, write, j)
		if align:
			write.align_stream(4)


"""



from enum import IntEnum


class AudioType(IntEnum):
    UNKNOWN = (0,)
    ACC = (1,)
    AIFF = (2,)
    IT = (10,)
    MOD = (12,)
    MPEG = (13,)
    OGGVORBIS = (14,)
    S3M = (17,)
    WAV = (20,)
    XM = (21,)
    XMA = (22,)
    VAG = (23,)
    AUDIOQUEUE = 24


class AudioCompressionFormat(IntEnum):
    PCM = (0,)
    Vorbis = (1,)
    ADPCM = (2,)
    MP3 = (3,)
    VAG = (4,)
    HEVAG = (5,)
    XMA = (6,)
    AAC = (7,)
    GCADPCM = (8,)
    ATRAC9 = 9


AUDIO_TYPE_EXTEMSION = {
    AudioType.ACC: ".m4a",
    AudioType.AIFF: ".aif",
    AudioType.IT: ".it",
    AudioType.MOD: ".mod",
    AudioType.MPEG: ".mp3",
    AudioType.OGGVORBIS: ".ogg",
    AudioType.S3M: ".s3m",
    AudioType.WAV: ".wav",
    AudioType.XM: ".xm",
    AudioType.XMA: ".wav",
    AudioType.VAG: ".vag",
    AudioType.AUDIOQUEUE: ".fsb",
    AudioCompressionFormat.PCM: ".fsb",
    AudioCompressionFormat.Vorbis: ".fsb",
    AudioCompressionFormat.ADPCM: ".fsb",
    AudioCompressionFormat.MP3: ".fsb",
    AudioCompressionFormat.VAG: ".vag",
    AudioCompressionFormat.HEVAG: ".vag",
    AudioCompressionFormat.XMA: ".wav",
    AudioCompressionFormat.AAC: ".m4a",
    AudioCompressionFormat.GCADPCM: ".fsb",
    AudioCompressionFormat.ATRAC9: ".at9",
}



from .EditorExtension import EditorExtension
from .PPtr import PPtr


class GameObject(EditorExtension):
    """
    public PPtr<Component>[] m_Components;
    public string m_Name;
    public Transform m_Transform;
    public MeshRenderer m_MeshRenderer;
    public MeshFilter m_MeshFilter;
    public SkinnedMeshRenderer m_SkinnedMeshRenderer;
    public Animator m_Animator;
    public Animation m_Animation;
    """

    def __init__(self, reader):
        super().__init__(reader=reader)
        component_size = reader.read_int()
        self.components = []
        for i in range(component_size):
            if self.version[0] < 5 or (self.version[0] == 5 and self.version[1] < 5):
                first = reader.read_int()
            self.components.append(PPtr(reader))
        self.layer = reader.read_int()
        self.name = reader.read_aligned_string()



from .AnimationClip import ValueArrayConstant
from .PPtr import PPtr
from .RuntimeAnimatorController import RuntimeAnimatorController
from ..math import Vector3


class AnimatorController(RuntimeAnimatorController):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_ControllerSize = reader.read_u_int()
        self.m_Controller = ControllerConstant(reader)
        tosSize = reader.read_int()
        self.m_TOS = {}
        for _ in range(tosSize):
            key = reader.read_u_int()
            self.m_TOS[key] = reader.read_aligned_string()

        numClips = reader.read_int()
        self.m_AnimationClips = [PPtr(reader) for _ in range(numClips)]


class HumanPoseMask:
    def __init__(self, reader):
        version = reader.version
        self.word0 = reader.read_u_int()
        self.word1 = reader.read_u_int()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 2):  # 5.2 and up
            self.word2 = reader.read_u_int()


class SkeletonMaskElement:
    def __init__(self, reader):
        self.m_PathHash = reader.read_u_int()
        self.m_Weight = reader.read_float()


class SkeletonMask:
    def __init__(self, reader):
        numElements = reader.read_int()
        self.m_Data = [SkeletonMaskElement(reader) for _ in range(numElements)]


class LayerConstant:
    def __init__(self, reader):
        version = reader.version
        self.m_StateMachineIndex = reader.read_u_int()
        self.m_StateMachineMotionSetIndex = reader.read_u_int()
        self.m_BodyMask = HumanPoseMask(reader)
        self.m_SkeletonMask = SkeletonMask(reader)
        self.m_Binding = reader.read_u_int()
        self.m_LayerBlendingMode = reader.read_int()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 2):  # 4.2 and up
            self.m_DefaultWeight = reader.read_float()
        self.m_IKPass = reader.read_boolean()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 2):  # 4.2 and up
            self.m_SyncedLayerAffectsTiming = reader.read_boolean()
        reader.align_stream()


class ConditionConstant:
    def __init__(self, reader):
        self.m_ConditionMode = reader.read_u_int()
        self.m_EventID = reader.read_u_int()
        self.m_EventThreshold = reader.read_float()
        self.m_ExitTime = reader.read_float()


class TransitionConstant:
    def __init__(self, reader):
        version = reader.version

        numConditions = reader.read_int()
        self.m_ConditionConstantArray = [
            ConditionConstant(reader) for _ in range(numConditions)
        ]

        self.m_DestinationState = reader.read_u_int()
        if version[0] >= 5:  # 5.0 and up
            self.m_FullPathID = reader.read_u_int()

        self.m_ID = reader.read_u_int()
        self.m_UserID = reader.read_u_int()
        self.m_TransitionDuration = reader.read_float()
        self.m_TransitionOffset = reader.read_float()
        if version[0] >= 5:  # 5.0 and up
            self.m_ExitTime = reader.read_float()
            self.m_HasExitTime = reader.read_boolean()
            self.m_HasFixedDuration = reader.read_boolean()
            reader.align_stream()
            self.m_InterruptionSource = reader.read_int()
            self.m_OrderedInterruption = reader.read_boolean()
        else:
            self.m_Atomic = reader.read_boolean()

        if version[0] > 4 or (version[0] == 4 and version[1] >= 5):  # 4.5 and up
            self.m_CanTransitionToSelf = reader.read_boolean()

        reader.align_stream()


class LeafInfoConstant:
    def __init__(self, reader):
        self.m_IDArray = reader.read_u_int_array()
        self.m_IndexOffset = reader.read_u_int()


class MotionNeighborList:
    def __init__(self, reader):
        self.m_NeighborArray = reader.read_u_int_array()


class Blend2dDataConstant:
    def __init__(self, reader):
        self.m_ChildPositionArray = reader.read_vector2_array()
        self.m_ChildMagnitudeArray = reader.read_float_array()
        self.m_ChildPairVectorArray = reader.read_vector2_array()
        self.m_ChildPairAvgMagInvArray = reader.read_float_array()

        numNeighbours = reader.read_int()
        self.m_ChildNeighborListArray = [
            MotionNeighborList(reader) for _ in range(numNeighbours)
        ]


class Blend1dDataConstant:  # wrong labeled:
    def __init__(self, reader):
        self.m_ChildThresholdArray = reader.read_float_array()


class BlendDirectDataConstant:
    def __init__(self, reader):
        self.m_ChildBlendEventIDArray = reader.read_u_int_array()
        self.m_NormalizedBlendValues = reader.read_boolean()
        reader.align_stream()


class BlendTreeNodeConstant:
    def __init__(self, reader):
        version = reader.version

        if version[0] > 4 or (version[0] == 4 and version[1] >= 1):  # 4.1 and up
            self.m_BlendType = reader.read_u_int()
        self.m_BlendEventID = reader.read_u_int()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 1):  # 4.1 and up
            self.m_BlendEventYID = reader.read_u_int()
        self.m_ChildIndices = reader.read_u_int_array()
        if version[0] < 4 or (version[0] == 4 and version[1] < 1):  # 4.1 down
            self.m_ChildThresholdArray = reader.read_float_array()

        if version[0] > 4 or (version[0] == 4 and version[1] >= 1):  # 4.1 and up
            self.m_Blend1dData = Blend1dDataConstant(reader)
            self.m_Blend2dData = Blend2dDataConstant(reader)

        if version[0] >= 5:  # 5.0 and up
            self.m_BlendDirectData = BlendDirectDataConstant(reader)

        self.m_ClipID = reader.read_u_int()
        if version[0] == 4 and version[1] >= 5:  # 4.5 - 5.0
            self.m_ClipIndex = reader.read_u_int()

        self.m_Duration = reader.read_float()

        if (
            version[0] > 4
            or (version[0] == 4 and version[1] > 1)
            or (version[0] == 4 and version[1] == 1 and version[2] >= 3)
        ):  # 4.1.3 and up
            self.m_CycleOffset = reader.read_float()
            self.m_Mirror = reader.read_boolean()
            reader.align_stream()


class BlendTreeConstant:
    def __init__(self, reader):
        version = reader.version

        numNodes = reader.read_int()
        self.m_NodeArray = [BlendTreeNodeConstant(reader) for _ in range(numNodes)]

        if version[0] < 4 or (version[0] == 4 and version[1] < 5):  # 4.5 down
            self.m_BlendEventArrayConstant = ValueArrayConstant(reader)


class StateConstant:
    def __init__(self, reader):
        version = reader.version

        numTransistions = reader.read_int()
        self.m_TransitionConstantArray = [
            TransitionConstant(reader) for _ in range(numTransistions)
        ]

        self.m_BlendTreeConstantIndexArray = reader.read_int_array()

        if version[0] < 5 or (version[0] == 5 and version[1] < 2):  # 5.2 down
            numInfos = reader.read_int()
            self.m_LeafInfoArray = [LeafInfoConstant(reader) for _ in range(numInfos)]

        numBlends = reader.read_int()
        self.m_BlendTreeConstantArray = [
            BlendTreeConstant(reader) for _ in range(numBlends)
        ]

        self.m_NameID = reader.read_u_int()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_PathID = reader.read_u_int()
        if version[0] >= 5:  # 5.0 and up
            self.m_FullPathID = reader.read_u_int()

        self.m_TagID = reader.read_u_int()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 1):  # 5.1 and up
            self.m_SpeedParamID = reader.read_u_int()
            self.m_MirrorParamID = reader.read_u_int()
            self.m_CycleOffsetParamID = reader.read_u_int()

        if version[0] > 2017 or (
            version[0] == 2017 and version[1] >= 2
        ):  # 2017.2 and up
            self.m_TimeParamID = reader.read_u_int()

        self.m_Speed = reader.read_float()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 1):  # 4.1 and up
            self.m_CycleOffset = reader.read_float()
        self.m_IKOnFeet = reader.read_boolean()
        if version[0] >= 5:  # 5.0 and up
            self.m_WriteDefaultValues = reader.read_boolean()

        self.m_Loop = reader.read_boolean()
        if version[0] > 4 or (version[0] == 4 and version[1] >= 1):  # 4.1 and up
            self.m_Mirror = reader.read_boolean()

        reader.align_stream()


class SelectorTransitionConstant:
    def __init__(self, reader):
        self.m_Destination = reader.read_u_int()

        numConditions = reader.read_int()
        self.m_ConditionConstantArray = [
            ConditionConstant(reader) for _ in range(numConditions)
        ]


class SelectorStateConstant:
    def __init__(self, reader):
        numTransitions = reader.read_int()
        self.m_TransitionConstantArray = [
            SelectorTransitionConstant(reader) for _ in range(numTransitions)
        ]
        self.m_FullPathID = reader.read_u_int()
        self.m_isEntry = reader.read_boolean()
        reader.align_stream()


class StateMachineConstant:
    def __init__(self, reader):
        version = reader.version

        numStates = reader.read_int()
        self.m_StateConstantArray = [StateConstant(reader) for _ in range(numStates)]

        numAnyStates = reader.read_int()
        self.m_AnyStateTransitionConstantArray = [
            TransitionConstant(reader) for _ in range(numAnyStates)
        ]

        if version[0] >= 5:  # 5.0 and up
            numSelectors = reader.read_int()
            self.m_SelectorStateConstantArray = [
                SelectorStateConstant(reader) for _ in range(numSelectors)
            ]

        self.m_DefaultState = reader.read_u_int()
        self.m_MotionSetCount = reader.read_u_int()


class ValueArray:
    def __init__(self, reader):
        version = reader.version

        if version[0] < 5 or (version[0] == 5 and version[1] < 5):  # 5.5 down
            self.m_BoolValues = reader.read_boolean_array()
            reader.align_stream()
            self.m_IntValues = reader.read_int_array()
            self.m_FloatValues = reader.read_float_array()

        if version[0] < 4 or (version[0] == 4 and version[1] < 3):  # 4.3 down
            self.m_VectorValues = reader.read_vector4_array()
        else:
            numPosValues = reader.read_int()
            self.m_PositionValues = [
                reader.read_vector3()
                if (version[0] > 5 or (version[0] == 5 and version[1] >= 4))
                else Vector3(reader.read_vector4())  # 5.4 and up
                for _ in range(numPosValues)
            ]

            self.m_QuaternionValues = reader.read_vector4_array()

            numScaleValues = reader.read_int()
            self.m_ScaleValues = [
                reader.read_vector3()
                if (version[0] > 5 or (version[0] == 5 and version[1] >= 4))
                else Vector3(reader.read_vector4())  # 5.4 and up
                for _ in range(numScaleValues)
            ]

            if version[0] > 5 or (version[0] == 5 and version[1] >= 5):  # 5.5 and up
                self.m_FloatValues = reader.read_float_array()
                self.m_IntValues = reader.read_int_array()
                self.m_BoolValues = reader.read_boolean_array()
                reader.align_stream()


class ControllerConstant:
    def __init__(self, reader):
        numLayers = reader.read_int()
        self.m_LayerArray = [LayerConstant(reader) for _ in range(numLayers)]

        numStates = reader.read_int()
        self.m_StateMachineArray = [
            StateMachineConstant(reader) for _ in range(numStates)
        ]

        self.m_Values = ValueArrayConstant(reader)
        self.m_DefaultValues = ValueArray(reader)



from .NamedObject import NamedObject
from ..helpers.ResourceReader import get_resource_data


class VideoClip(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.original_path = reader.read_aligned_string()
        self.proxy_width = reader.read_u_int()
        self.proxy_height = reader.read_u_int()
        self.width = reader.read_u_int()
        self.height = reader.read_u_int()
        if self.version[0] > 2017 or (
            self.version[0] == 2017 and self.version[1] >= 2
        ):  # 2017.2 and up
            self.pixel_aspec_ratio_num = reader.read_u_int()
            self.pixel_aspec_ratio_den = reader.read_u_int()
        self.frame_rate = reader.read_double()
        self.frame_count = reader.read_u_long()
        self.format = reader.read_int()
        self.audio_channel_count = reader.read_u_short_array()
        reader.align_stream()
        self.audio_sample_rate = reader.read_u_int_array()
        self.audio_language = reader.read_string_array()
        # StreamedResource m_ExternalResources
        source = reader.read_aligned_string()
        offset = reader.read_u_long()
        size = reader.read_u_long()
        self.has_split_alpha = reader.read_boolean()
        if source:
            self.VideoData = get_resource_data(source, self.assets_file, offset, size)
        else:
            self.VideoData = reader.read(size)



from .Color import Color
from .Vector2 import Vector2
from .Vector3 import Vector3
from .Vector4 import Vector4
from .Matrix4x4 import Matrix4x4
from .Quaternion import Quaternion
from .Rectangle import Rectangle



from .EditorExtension import EditorExtension
from ..streams import EndianBinaryWriter


class NamedObject(EditorExtension):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.reader.reset()
        self.name = self.reader.read_aligned_string()

    def save(self, writer: EndianBinaryWriter):
        super().save(writer)
        writer.write_aligned_string(self.name)



import io
import os
from zipfile import ZipFile

from . import files
from .enums import FileType
from .helpers import ImportHelper


class Environment:
    files: dict
    resources: dict
    container: dict
    assets: dict
    path: str

    def __init__(self, *args):
        self.files = {}
        self.container = {}
        self.assets = {}
        self.resources = {}
        self.path = "."

        if args:
            for arg in args:
                if isinstance(arg, str):
                    if os.path.isfile(arg):
                        if os.path.splitext(arg)[-1] in [".apk", ".zip"]:
                            self.load_zip_file(arg)
                        else:
                            self.path = os.path.dirname(arg)
                            self.load_file(arg)
                    elif os.path.isdir(arg):
                        self.path = arg
                        self.load_folder(arg)
                else:
                    self.path = None
                    self.load_file(data=arg)

        if len(self.files) == 1:
            self.file = list(self.files.values())[0]

    def load_files(self, files: list):
        """Loads all files (list) into the AssetsManager and merges .split files for common usage."""
        path = os.path.dirname(files[0])
        ImportHelper.merge_split_assets(path)
        to_read_file = ImportHelper.processing_split_files(files)
        self.load(to_read_file)

    def load_folder(self, path: str):
        """Loads all files in the given path and its subdirs into the AssetsManager."""
        ImportHelper.merge_split_assets(path, True)
        files = ImportHelper.list_all_files(path)
        to_read_file = ImportHelper.processing_split_files(files)
        self.load(to_read_file)

    def load(self, files: list):
        """Loads all files into the AssetsManager."""
        # for f in files:
        #    self.import_files[os.path.basename(f)] = f

        # self.Progress.reset()
        # use a for loop because list size can change
        # for i, f in enumerate(self.import_files.values()):
        for f in files:
            self.load_file(f)
            # self.Progress.report(i + 1, len(self.import_files))

    def load_file(self, full_name: str = "", data=None):
        typ, reader = ImportHelper.check_file_type(data if data else full_name)
        if not full_name:
            full_name = str(data[:256])
        if typ == FileType.AssetsFile:
            self.files[full_name] = files.SerializedFile(reader, self)
            self.assets[full_name] = self.files[full_name]
        elif typ == FileType.BundleFile:
            self.files[full_name] = files.BundleFile(reader, self)
        elif typ == FileType.WebFile:
            self.files[full_name] = files.WebFile(reader, self)
        elif typ == FileType.ZIP:
            self.load_zip_file(reader.stream)
        elif typ == FileType.ResourceFile:
            self.resources[os.path.basename(full_name)] = reader

    def load_zip_file(self, value):
        buffer = None
        if isinstance(value, str) and os.path.exists(value):
            buffer = open(value, "rb")
        elif isinstance(value, (bytes, bytearray)):
            buffer = ZipFile(io.BytesIO(value))
        elif isinstance(value, (io.BufferedReader, io.BufferedIOBase)):
            buffer = value

        z = ZipFile(buffer)

        for path in z.namelist():
            self.load_file(path, z.open(path).read())

    @property
    def objects(self):
        def search(item):
            ret = []
            #print(type(item))
            if not isinstance(item, Environment) and getattr(item, "objects", None):
                # serialized file
                return [val for val in item.objects.values()]

            elif getattr(item, "files", None):  # WebBundle and BundleFile
                # bundle
                for item in item.files.values():
                    ret.extend(search(item))
                return ret

            return ret

        return search(self)



from ..streams import EndianBinaryReader, EndianBinaryWriter


class PPtr:
    def __init__(self, reader: EndianBinaryReader):
        self.index = -2
        self.file_id = reader.read_int()
        self.path_id = reader.read_int() if reader.version2 < 14 else reader.read_long()
        self.assets_file = reader.assets_file

    def __getattr__(self, key):
        # get manager
        manager = None
        if self.file_id == 0:
            manager = self.assets_file

        elif self.file_id > 0 and self.file_id - 1 < len(self.assets_file.externals):
            if self.index == -2:
                external_name = self.assets_file.externals[self.file_id - 1].name

                files = self.assets_file.parent.files
                if external_name not in files:
                    external_name = external_name.upper()
                manager = self.assets_file.parent.files[external_name]

        if manager and self.path_id in manager.objects:
            self = manager.objects[self.path_id]
            return getattr(self, key)

        raise NotImplementedError("PPtr")

        """
				if (!assetsFileIndexCache.TryGetValue(name, out index))
				{
					index = assetsFileList.FindIndex(x => x.upperFileName == name);
					assetsFileIndexCache.Add(name, index);
				}
			}

			if (index >= 0)
			{
				result = assetsFileList[index];
				return true;
			}
		}
		if self.path_id in self.assets_file.objects:
			return self.assets_file.objects[self.path_id]
		else:
			return super(PPtr, self).__new__(self)
	"""

    def __repr__(self):
        return self.__class__.__name__

    def __bool__(self):
        return False


def save_ptr(obj, writer: EndianBinaryWriter):
    if isinstance(obj, PPtr):
        writer.write_int(obj.file_id)
        writer.write_int(obj.path_id)
    else:
        writer.write_int(0)  # it's usually 0......
        writer.write_int(obj.path_id)


"""	
	def TryGetAssetsFile(self):
		result = None
		if m_FileID == 0:
			return self.assets_file

		if m_FileID > 0 and m_FileID - 1 < len(self.assets_file.externals):
			var assetsManager = assetsFile.assetsManager
			var assetsFileList = assetsManager.assetsFileList
			var assetsFileIndexCache = assetsManager.assetsFileIndexCache

			if self.index == -2:
				external_name = self.assets_file.eternals[self.file_id - 1].filename

				
				if (!assetsFileIndexCache.TryGetValue(name, out index))
				{
					index = assetsFileList.FindIndex(x => x.upperFileName == name);
					assetsFileIndexCache.Add(name, index);
				}
			}

			if (index >= 0)
			{
				result = assetsFileList[index];
				return true;
			}
		}

		return false;
	}

		public bool TryGet(out T result)
		{
			if (TryGetAssetsFile(out var sourceFile))
			{
				if (sourceFile.Objects.TryGetValue(m_PathID, out var obj))
				{
					if (obj is T variable)
					{
						result = variable;
						return true;
					}
				}
			}

			result = null;
			return false;
		}

		public bool TryGet<T2>(out T2 result) where T2 : Object
		{
			if (TryGetAssetsFile(out var sourceFile))
			{
				if (sourceFile.Objects.TryGetValue(m_PathID, out var obj))
				{
					if (obj is T2 variable)
					{
						result = variable;
						return true;
					}
				}
			}

			result = null;
			return false;
		}

		public void Set(T m_Object)
		{
			var name = m_Object.assetsFile.upperFileName;
			if (string.Equals(assetsFile.upperFileName, name, StringComparison.Ordinal))
			{
				m_FileID = 0;
			}
			else
			{
				m_FileID = assetsFile.m_Externals.FindIndex(x => string.Equals(x.fileName, name, StringComparison.OrdinalIgnoreCase));
				if (m_FileID == -1)
				{
					assetsFile.m_Externals.Add(new FileIdentifier
					{
						fileName = m_Object.assetsFile.fileName
					});
					m_FileID = assetsFile.m_Externals.Count;
				}
				else
				{
					m_FileID += 1;
				}
			}

			var assetsManager = assetsFile.assetsManager;
			var assetsFileList = assetsManager.assetsFileList;
			var assetsFileIndexCache = assetsManager.assetsFileIndexCache;

			if (!assetsFileIndexCache.TryGetValue(name, out index))
			{
				index = assetsFileList.FindIndex(x => x.upperFileName == name);
				assetsFileIndexCache.Add(name, index);
			}

			m_PathID = m_Object.m_PathID;
		}

		public bool IsNull => m_PathID == 0 || m_FileID < 0;
	}
}
"""



from .BundleFile import BundleFile
from .File import File
from .SerializedFile import SerializedFile
from ..enums import FileType
from ..helpers import CompressionHelper, ImportHelper
from ..streams import EndianBinaryReader, EndianBinaryWriter


class WebFile(File):
    """A package which can hold other WebFiles, Bundles and SerialiedFiles.
    It may be compressed via gzip or brotli.

    files -- list of all files in the WebFile
    """

    def __init__(self, reader: EndianBinaryReader, environment=None):
        """Constructor Method
        """
        self.files = {}

        # check compression
        magic = reader.read_bytes(2)
        reader.Position = 0

        if magic == CompressionHelper.GZIP_MAGIC:
            self.compression = "gzip"
            data = CompressionHelper.decompress_gzip(reader.bytes)
            reader = EndianBinaryReader(data, endian="<")
        else:
            reader.Position = 0x20
            magic = reader.read_bytes(6)
            reader.Position = 0
            if CompressionHelper.BROTLI_MAGIC == magic:
                self.compression = "brotli"
                data = CompressionHelper.decompress_brotli(reader.bytes)
                reader = EndianBinaryReader(data, endian="<")
            else:
                self.compression = "None"
                reader.endian = "<"

        # signature check
        signature = reader.read_string_to_null()
        if signature != "UnityWebData1.0":
            return
        self.signature = signature

        # read header -> contains file headers
        head_length = reader.read_int()

        files = []
        while reader.Position < head_length:
            offset = reader.read_int()
            length = reader.read_int()
            path_length = reader.read_int()
            name = reader.read_bytes(path_length).decode("utf8")
            files.append((name, offset, length))

        # read file data and convert it
        for name, offset, length in files:
            reader.Position = offset
            data = reader.read(length)
            typ, item = ImportHelper.check_file_type(data)

            if typ == FileType.BundleFile:
                item = BundleFile(item, environment)
            elif typ == FileType.WebFile:
                item = WebFile(item, environment)

            if typ == FileType.AssetsFile:
                # pre-check if resource file
                if name.endswith((".resS", ".resource", ".config", ".xml", ".dat")):
                    environment.resources[name] = item
                else:
                    # try to load the file as serialized file
                    try:
                        item = SerializedFile(item, environment)
                        environment.assets[name] = item
                    except ValueError:
                        environment.resources[name] = item

            self.files[name] = item
            item.parent = self

    def save(
        self,
        files: dict = None,
        compression: str = None,
        signature: str = "UnityWebData1.0",
    ) -> bytes:
        # solve defaults
        if not files:
            files = self.files
        if not compression:
            compression = self.compression

        # get raw data
        files = {
            name: f.bytes if isinstance(f, EndianBinaryReader) else f.save()
            for name, f in files.items()
        }

        # create writer
        writer = EndianBinaryWriter(endian="<")
        # signature
        writer.write_string_to_null(signature)

        # data offset
        offset = sum(
            [
                writer.Position,  # signature
                sum(
                    len(path.encode("utf8")) for path in files.keys()
                ),  # path of each file
                4 * 3 * len(files),  # 3 ints per file
                4,  # offset int
            ]
        )

        writer.write_int(offset)

        # 1. file headers
        for name, data in files.items():
            # offset
            writer.write_int(offset)
            # length
            length = len(data)
            writer.write_int(length)
            offset += length
            # path
            enc_path = name.encode("utf8")
            writer.write_int(len(enc_path))
            writer.write(enc_path)

        # 2. file data
        for data in files.values():
            writer.write(data)

        if compression == "gzip":
            return CompressionHelper.compress_gzip(writer.bytes)
        elif compression == "brotli":
            return CompressionHelper.compress_brotli(writer.bytes)
        else:
            return writer.bytes



from .AnimationClip import xform
from .NamedObject import NamedObject


class Avatar(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_AvatarSize = reader.read_u_int()
        self.m_Avatar = AvatarConstant(reader)

        numTOS = reader.read_int()
        self.m_TOS = {}
        for _ in range(numTOS):
            key = reader.read_u_int()
            self.m_TOS[key] = reader.read_aligned_string()

    # HumanDescription m_HumanDescription 2019 and up

    def FindBonePath(self, hash):
        return self.m_TOS[hash]


class Node:
    def __init__(self, reader):
        self.m_ParentId = reader.read_int()
        self.m_AxesId = reader.read_int()


class Limit:
    def __init__(self, reader):
        version = reader.version
        if version[0] > 5 or (version[0] == 5 and version[1] >= 4):  # 5.4 and up
            self.m_Min = reader.read_vector3()
            self.m_Max = reader.read_vector3()
        else:
            self.m_Min = reader.read_vector4()
            self.m_Max = reader.read_vector4()


class Axes:
    def __init__(self, reader):
        version = reader.version
        self.m_PreQ = reader.read_vector4()
        self.m_PostQ = reader.read_vector4()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 4):  # 5.4 and up
            self.m_Sgn = reader.read_vector3()
        else:
            self.m_Sgn = reader.read_vector4()
        self.m_Limit = Limit(reader)
        self.m_Length = reader.read_float()
        self.m_Type = reader.read_u_int()


class Skeleton:
    def __init__(self, reader):
        numNodes = reader.read_int()
        self.m_Node = [Node(reader) for _ in range(numNodes)]

        self.m_ID = reader.read_u_int_array()

        numAxes = reader.read_int()
        self.m_AxesArray = [Axes(reader) for _ in range(numAxes)]


class SkeletonPose:
    def __init__(self, reader):
        numXforms = reader.read_int()
        self.m_X = [xform(reader) for _ in range(numXforms)]


class Hand:
    def __init__(self, reader):
        self.m_HandBoneIndex = reader.read_int_array()


class Handle:
    def __init__(self, reader):
        self.m_X = xform(reader)
        self.m_ParentHumanIndex = reader.read_u_int()
        self.m_ID = reader.read_u_int()


class Collider:
    def __init__(self, reader):
        self.m_X = xform(reader)
        self.m_Type = reader.read_u_int()
        self.m_XMotionType = reader.read_u_int()
        self.m_YMotionType = reader.read_u_int()
        self.m_ZMotionType = reader.read_u_int()
        self.m_MinLimitX = reader.read_float()
        self.m_MaxLimitX = reader.read_float()
        self.m_MaxLimitY = reader.read_float()
        self.m_MaxLimitZ = reader.read_float()


class Human:
    def __init__(self, reader):
        version = reader.version
        self.m_RootX = xform(reader)
        self.m_Skeleton = Skeleton(reader)
        self.m_SkeletonPose = SkeletonPose(reader)
        self.m_LeftHand = Hand(reader)
        self.m_RightHand = Hand(reader)

        if version[0] < 2018 or (version[0] == 2018 and version[1] < 2):  # 2018.2 down
            numHandles = reader.read_int()
            self.m_Handles = [Handle(reader) for _ in range(numHandles)]
            numColliders = reader.read_int()
            self.m_ColliderArray = [Collider(reader) for _ in range(numColliders)]
        self.m_HumanBoneIndex = reader.read_int_array()
        self.m_HumanBoneMass = reader.read_float_array()

        if version[0] < 2018 or (version[0] == 2018 and version[1] < 2):  # 2018.2 down
            self.m_ColliderIndex = reader.read_int_array()

        self.m_Scale = reader.read_float()
        self.m_ArmTwist = reader.read_float()
        self.m_ForeArmTwist = reader.read_float()
        self.m_UpperLegTwist = reader.read_float()
        self.m_LegTwist = reader.read_float()
        self.m_ArmStretch = reader.read_float()
        self.m_LegStretch = reader.read_float()
        self.m_FeetSpacing = reader.read_float()
        self.m_HasLeftHand = reader.read_boolean()
        self.m_HasRightHand = reader.read_boolean()
        if version[0] > 5 or (version[0] == 5 and version[1] >= 2):  # 5.2 and up
            self.m_HasTDoF = reader.read_boolean()
        reader.align_stream()


class AvatarConstant:
    def __init__(self, reader):
        version = reader.version
        self.m_AvatarSkeleton = Skeleton(reader)
        self.m_AvatarSkeletonPose = SkeletonPose(reader)

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_DefaultPose = SkeletonPose(reader)
            self.m_SkeletonNameIDArray = reader.read_u_int_array()

        self.m_Human = Human(reader)

        self.m_HumanSkeletonIndexArray = reader.read_int_array()

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_HumanSkeletonReverseIndexArray = reader.read_int_array()

        self.m_RootMotionBoneIndex = reader.read_int()
        self.m_RootMotionBoneX = xform(reader)

        if version[0] > 4 or (version[0] == 4 and version[1] >= 3):  # 4.3 and up
            self.m_RootMotionSkeleton = Skeleton(reader)
            self.m_RootMotionSkeletonPose = SkeletonPose(reader)

            self.m_RootMotionSkeletonIndexArray = reader.read_int_array()
